import copy
from collections import OrderedDict

from .TreeItem import TreeItem
from .DictTree import DictTree

class TreeModel(object):
    """
    This class indexes a DictTree with a set of TreeItems.
    TreeItems keep track of their lineage in the DictTree,
    and can be modified for additional functionality
    in subclasses of TreeModel by adding TreeItem.flags.
    """

    def __init__(self,default_flags={}):
        super(TreeModel,self).__init__()
        # a TreeItem with no parent is the root of the tree
        self._root_item = TreeItem(None,'ROOT')
        # the underlying data are stored in a DictTree. 
        self._tree = DictTree()
        # any TreeItems will be given these default_flags,
        # unless subclasses override TreeModel.create_tree_item()
        self.default_flags = default_flags

    def __getitem__(self,uri):
        return self.get_from_uri(uri)

    def __setitem__(self,uri,val):
        self.set_item(uri,val)
    
    def n_children(self,parent_uri=''):
        itm = self.get_from_uri(parent_uri)
        return itm.n_children()

    def set_item(self,itm_uri,itm_data=None):
        #try:
        if '.' in itm_uri:
            parent_uri = itm_uri[:itm_uri.rfind('.')]
            parent_itm = self.get_from_uri(parent_uri)
            itm_tag = itm_uri.split('.')[-1]
        else:
            parent_itm = self._root_item
            itm_tag = itm_uri
        # cast the itm_data as an embedded dict for tree-storage 
        treedata = self.build_tree(itm_data)
        # add TreeItems to index the new TreeModel content 
        self.tree_update(parent_itm,itm_tag,treedata)
        # store the data 
        self._tree.set_uri(itm_uri,itm_data)
        #except Exception as ex:
        #    msg = str('[{}] Encountered an error while trying to set uri {}: \n'
        #    .format(__name__,itm_uri) + ex.message)
        #    raise KeyError(msg) 

    def remove_item(self,itm_uri):
        self._tree.delete_uri(itm_uri)
        # remove the corresponding subtree or TreeItem.
        itm = self.get_from_uri(itm_uri)
        parent_itm = itm.parent 
        child_keys = [c.tag for c in parent_itm.children]
        rm_row = child_keys.index(itm.tag)
        parent_itm.children.pop(rm_row)

    def tree_update(self,parent_itm,itm_tag,itm_data):
        """
        Update the tree structure 
        rooted at parent_itm.children[itm_tag],
        such that TreeItems get built 
        to index all of the items in itm_data 
        that are supported by self.build_tree().
        Assume build_tree was called on itm_data
        before passing it as an argument,
        so only need to recurse if itm_data is a dict.
        """
        child_keys = [c.tag for c in parent_itm.children]
        if itm_tag in child_keys:
            # Find the row of existing itm_tag under parent,
            itm_row = child_keys.index(itm_tag)
            itm = parent_itm.children[itm_row]
            # Remove any grandchildren that do not represent the new data
            new_keys = []
            if isinstance(itm_data,dict):
                new_keys = itm_data.keys()
            for gc_row in range(itm.n_children())[::-1]:
                gc_itm = itm.children[gc_row]
                if not gc_itm.tag in new_keys:
                    itm.children.pop(gc_row) 
        else:
            # Put a new TreeItem at the end row
            itm_row = parent_itm.n_children()
            itm = self.create_tree_item(parent_itm,itm_tag)
            parent_itm.children.insert(itm_row,itm)
        # If needed, recurse on itm_data.
        if isinstance(itm_data,dict):
            for tag,val in itm_data.items():
                self.tree_update(itm,tag,val)

    def build_tree(self,x):
        """
        TreeModel.build_tree is called on some object x
        before x is stored in the tree.
        For subclasses of TreeModel to build tree data
        for data types other than dicts and lists,
        build_tree should be reimplemented.
        If data types other than dicts and lists have child items 
        that should be accessible by TreeModel uris,
        they should implement __getitem__(tag).
        """
        if isinstance(x,dict):
            d = OrderedDict()
            for k,v in x.items():
                d[k] = self.build_tree(v) 
            return d
        elif isinstance(x,list):
            d = OrderedDict(zip([str(i) for i in range(len(x))],x)) 
            for k,v in d.items():
                d[k] = self.build_tree(v)
            return d
        else:
            return x

    def get_from_uri(self,uri):
        try:
            path = uri.split('.')
            itm = self._root_item
            for k in path[:-1]:
                child_keys = [c.tag for c in itm.children]
                itm = itm.children[child_keys.index(k)]
            k = path[-1]
            if k:
                child_keys = [c.tag for c in itm.children]
                return itm.children[child_keys.index(k)]
            elif k == '':
                return itm 
        except Exception as ex:
            msg = '\n[{}] Encountered an error while fetching uri {}\n'.format(__name__,uri)
            ex.message = msg + ex.message
            raise ex

    def get_data_from_uri(self,uri):
        return self._tree.get_from_uri(uri)

    def root_tags(self):
        return self._tree.root_keys()

    def tag_error_message(self,tag):
        return self._tree.tag_error_message(tag)

    def is_uri_valid(self,uri):
        return self._tree.is_uri_valid(uri)

    def is_tag_valid(self,tag):
        return self._tree.is_tag_valid(tag)

    def make_unique_uri(self,prefix):
        return self._tree.make_unique_uri(prefix)

    def contains_uri(self,uri):
        return self._tree.contains_uri(uri)

    def build_uri(self,itm):
        """
        Build a URI for TreeItem itm by combining the tags 
        of the lineage of itm, with '.' as a delimiter.
        """
        if itm == self._root_item:
            return ''
        else:
            uri = itm.tag
            while not itm.parent == self._root_item:
                itm = itm.parent
                uri = itm.tag+"."+uri
            return uri

    def create_tree_item(self,parent_itm,itm_tag):
        """
        Build a TreeItem for use in this tree.
        Reimplement create_tree_item() in subclasses of TreeModel
        to add features to TreeItems, such as default values for TreeItem.flags.
        TreeModel implementation returns TreeItem(parent_itm,itm_tag).
        """
        itm = TreeItem(parent_itm,itm_tag)
        itm.flags = copy.copy(self.default_flags)
        return itm



import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation
from ... import optools

class BgSubtractByTemperature(Operation):
    """
    Originally contributed by Amanda Fournier.  

    Find a background spectrum from a batch of background spectra,
    where the temperature of the background spectrum is as close as possible
    to the (input) temperature of the measured spectrum.
    Then subtract that background spectrum from the input spectrum.
    The measured and background spectra are expected to have the same domain.
    """
    
    def __init__(self):
        input_names = ['q_I_meas','T_meas','bg_batch_output','q_I_bg_key','T_bg_key']
        output_names = ['q_I_bgsub', 'T_bg', 'bg_factor']
        super(BgSubtractByTemperature, self).__init__(input_names, output_names)
        self.input_doc['q_I_meas'] = 'n-by-2 array of I(q) versus q'
        self.input_doc['T_meas'] = 'temperature as taken from the dict '\
            'produced by the detector header file'
        self.input_doc['bg_batch_output'] = 'the output (list of dicts) '\
            'of a batch of background spectra at different temperatures'
        self.input_doc['q_I_bg_key'] = 'the name of the bg_batch workflow output '\
            'containing n-by-2 array of q and I(q) for the background spectra '
        self.input_doc['T_bg_key'] = 'the name of the bg_batch workflow output '\
            'containing the background spectrum temperatures'
        self.output_doc['q_I_bgsub'] = 'n-by-2 array of q and '\
            'background-subracted intensity (I_meas-bg_factor*I_bg)'
        self.output_doc['T_bg'] = 'temperature of the subtracted background spectrum'
        self.output_doc['bg_factor'] = 'correction factor applied to background '\
            'before subtraction, to ensure positive background-subtracted intensities'
        self.input_type['q_I_meas'] = opmod.workflow_item
        self.input_type['T_meas'] = opmod.workflow_item
        self.input_type['bg_batch_output'] = opmod.workflow_item

    def run(self):
        q_I_meas = self.inputs['q_I_meas']
        T_meas = self.inputs['T_meas']
        qIkey = self.inputs['q_I_bg_key']
        Tkey = self.inputs['T_bg_key']
        bg_out = self.inputs['bg_batch_output']
        if q_I_meas is None or bg_out is None or not all([qIkey,Tkey]) or T_meas is None:
            return
        T_allbg = [d[Tkey] for d in bg_out]
        closest_T_idx = np.argmin(np.abs([T_bg - T_meas for T_bg in T_allbg]))
        T_bg = T_allbg[closest_T_idx]
        q_I_bg = bg_out[closest_T_idx][qIkey]
        if not all(q_I_meas[:,0] == q_I_bg[:,0]):
            msg = 'SPECTRUM AND BACKGROUND ON DIFFERENT q DOMAINS'
            raise ValueError(msg)
        bad_data = ( (q_I_meas[:,1] <= 0) 
            | (q_I_bg[:,1] <= 0) 
            | np.isnan(q_I_meas[:,1]) 
            | np.isnan(q_I_bg[:,1]) )
        I_floor = 0 
        #I_floor = 1E-6 * np.max(q_I_meas[:,1])
        bg_factor = np.min((q_I_meas[:,1][~bad_data]-I_floor) / q_I_bg[:,1][~bad_data])
        q_I_bgsub = np.array(q_I_meas)
        q_I_bgsub[:,1] = q_I_meas[:,1] - bg_factor * q_I_bg[:,1]

        self.outputs['q_I_bgsub'] = q_I_bgsub
        self.outputs['T_bg'] = T_bg 
        self.outputs['bg_factor'] = bg_factor





from __future__ import print_function
import unittest
import sys

import paws.api
from paws.core.operations.Operation import Operation

class TestOp(unittest.TestCase):

    def __init__(self,test_name,op_uri,paw):
        super(TestOp,self).__init__(test_name)
        self.op_uri = op_uri
        self.paw = paw

    def test_activate_op(self):
        print('testing {} ...'.format(self.op_uri),end=''); sys.stdout.flush()
        try:
            self.paw.activate_op(self.op_uri) 
        except ImportError as ex:
            msg = 'Caught ImportError: '\
            'missing packages for {}. '.format(self.op_uri)
            self.skipTest(msg)            

    def test_op(self):
        print('testing {} ...'.format(self.op_uri),end=''); sys.stdout.flush()
        op = self.paw._op_manager.get_data_from_uri(self.op_uri)
        op = op()
        self.assertIsInstance(op,Operation)
    
    def test_run(self):
        print('testing {} ...'.format(self.op_uri),end=''); sys.stdout.flush()
        op = self.paw._op_manager.get_data_from_uri(self.op_uri)
        op = op()
        op.load_defaults()
        op.run() 




import numpy as np

from .. import Operation as opmod 
from ..Operation import Operation

class LogLogZip(Operation):
    """
    Take the base-10 logarithm of two 1d arrays, then zip them together.
    Any elements with non-positive or nan values are removed. 
    """

    def __init__(self):
        input_names = ['x', 'y']
        output_names = ['logx_logy']
        super(LogLogZip, self).__init__(input_names, output_names)
        self.input_doc['x'] = '1d array'
        self.input_doc['y'] = '1d array, same size as x'
        self.output_doc['logx_logy'] = 'n x 2 array containing log(x) and log(y)'
        self.input_type['x'] = opmod.workflow_item
        self.input_type['y'] = opmod.workflow_item

    def run(self):
        x = self.inputs['x']
        y = self.inputs['y']
        if x is None or y is None:
            return
        # good_vals = elements for which both x and y have defined logarithm
        good_vals = ((x > 0) & (y > 0) & (~np.isnan(x)) & (~np.isnan(y)))
        xy = zip(np.log10(x[good_vals]), np.log10(y[good_vals]))
        self.outputs['logx_logy'] = np.array(xy)





import numpy as np
import pyFAI

from ... import Operation as opmod 
from ...Operation import Operation

class Integrate1d(Operation):
    """
    Integrate an image, given calibration parameters.

    Input image data (ndarray) and a dict of .poni format calibration parameters
    Output q, I(q) 
    """
    def __init__(self):
        input_names = ['image_data','poni_dict','fpolz']
        output_names = ['q','I','q_I']
        super(Integrate1d,self).__init__(input_names,output_names)
        self.input_doc['image_data'] = '2d array representing intensity for each pixel'
        self.input_doc['poni_dict'] = str( 'dict of calibration parameters; '
        + 'minimally including keys dist, poni1, poni2, rot1, rot2, rot3, pixel1, pixel2, wavelength;'
        + 'optionally including keys fpolz, detector, splineFile; '
        + 'same specifications as pyFAI .poni format calibration parameters')
        self.input_doc['fpolz'] = 'Polarization factor- default 1.0'
        self.input_type['image_data'] = opmod.workflow_item
        self.input_type['poni_dict'] = opmod.workflow_item
        self.inputs['fpolz'] = float(1.)
        self.output_doc['q'] = 'Scattering vector magnitude q in 1/Angstrom.'
        self.output_doc['I'] = 'Integrated intensity at q.'
        self.output_doc['q_I'] = 'q and I zipped together an a n-by-2 numpy array.'

    def run(self):
        img = self.inputs['image_data']
        pd = self.inputs['poni_dict']
        if img is None or pd is None:
            return
        p = pyFAI.AzimuthalIntegrator()
        p.setPyFAI(**pd)
        fpolz = self.inputs['fpolz']
        # use a mask to screen negative pixels
        # mask should be 1 for masked pixels, 0 for unmasked pixels
        msk = np.ones(img.shape)*(img <= 0)
        q, I_of_q = p.integrate1d(img, 1000, radial_range=(0.0005,1.0005), mask=msk, polarization_factor=fpolz, unit='q_A^-1')
        # save results to self.outputs
        self.outputs['q'] = q
        self.outputs['I'] = I_of_q
        self.outputs['q_I'] = np.array([q,I_of_q]).T




from collections import OrderedDict

from ... import Operation as opmod 
from ...Operation import Operation

import pyFAI

class ReadPONI(Operation):
    """
    Read in a dict of PyFAI PONI parameters.
    Input path to a .poni file representing a calibrated measurement geometry.
    """
    
    def __init__(self):
        input_names = ['file_path']
        output_names = ['poni_dict']
        super(ReadPONI,self).__init__(input_names,output_names)
        self.input_doc['poni_file'] = 'Path to a .poni file '\
            'describing a calibrated sample-detector geometry'
        self.output_doc['poni_dict'] = 'Dict of pyFAI calibration parameters'

    def run(self):
        fpath = self.inputs['file_path']
        g = pyFAI.geometry.Geometry()
        g.read(fpath)
        pdict = g.getPyFAI()
        self.outputs['poni_dict'] = pdict 




import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class LogY(Operation):
    """
    Take the base-10 logarithm of the second column
    of a n-by-2 array.
    """

    def __init__(self):
        input_names = ['x_y']
        output_names = ['x_logy']
        super(LogY, self).__init__(input_names, output_names)
        self.input_doc['x_y'] = 'n-by-2 array of x and y values'
        self.output_doc['x_logy'] = 'n-by-2 array of x and log_10(y) values'
        self.input_type['x_y'] = opmod.workflow_item

    def run(self):
        x_y = self.inputs['x_y']
        if x_y is None:
            return
        # good_vals = elements for which y has defined logarithm
        idx_ok = ((x_y[:,1] > 0) & (~np.isnan(x_y[:,1])))
        x_logy = np.zeros(x_y.shape)
        x_logy[idx_ok,1] = np.log10(x_y[idx_ok,1])
        x_logy[np.invert(idx_ok),1] = np.nan
        x_logy[:,0] = x_y[:,0]
        self.outputs['x_logy'] = x_logy 





from twisted.internet import task, reactor
from twisted.internet.protocol import ClientFactory
from twisted.protocols.basic import LineReceiver
from twisted.protocols.policies import TimeoutMixin

from .PawsPlugin import PawsPlugin
from ..operations import Operation as opmod

class TCPClientPlugin(PawsPlugin):

    def __init__(self):
        input_names = ['host','port']
        super(TCPClientPlugin,self).__init__(input_names)
        self.input_doc['host'] = 'string representing host name or IP address'
        self.input_doc['port'] = 'integer port number at which some server will be listening' 
       
    def start(self):
        self.host = self.inputs['host'] 
        self.port = self.inputs['port'] 
        self.client_factory = TCPClientFactory(TCPTestProtocol)
        # Set up the TCP connection with host and port, specify the factory
        reactor.connectTCP(self.host, self.port, self.client_factory)

    def stop(self):
        reactor.stop()
        # TODO: Anything else for cleaning up the connections.

    def description(self):
        desc = str('TCP Client Plugin for Paws: '
            + 'This is a TCP Client for general purpose and testing.')
        return desc

    def content(self):
        return {'inputs':self.inputs,'history':self.client_factory.history,'client_factory':self.client_factory}

    def send_text(self,txt):
        self.client_factory.cmd_list = [txt]
        reactor.run()                   

class TCPTestProtocol(LineReceiver):

    def __init__(self):
        """
        It is expected that this constructor will be called by a ClientFactory.
        The ClientFactory is expected to set protocol.factory to refer to itself
        after instantiating the protocol.
        The ClientFactory is expected to have a .history (list of strings).
        """
        #print "Protocol build"
        #self.delimiter = '\r\n'
        self.delimiter = '\n'
        self.cmdList = []

    def addCommand(self, cmd): 
        self.cmdList.append(cmd)

    def connectionMade(self):
        # Set up any objects specific to this Protocol.
        self.factory.history.append('CONNECTION MADE')
        #self.sendLineFromList() 

    def lineReceived(self, line):
        self.factory.history.append('RECEIVED: {}'.format(line))
 
    def connectionLost(self):
        # Dereference any objects specific to this Protocol.
        self.factory.history.append('CLIENT LOST CONNECTION')
        self.factory = None 
        self.cmdList = None 
    
    def send_lines(self):
        if self.cmdList == []:
            self.factory.history.append('COMMAND LIST EMPTY: LOSING CONNECTION')
            self.transport.loseConnection()
        else:
            line = self.cmdList.pop(0)
            #print "send: %s" % line
            self.factory.history.append('SEND: {}'.format(line))
            #LineReceiver.sendLine(self, line)
            self.sendLine(line)
            # call self.timeoutConnection after 5 seconds
            #self.setTimeout(5)

    #def timeoutConnection(self):
    #    #print "Connection timeout, communication aborted"
    #    self.factory.history.append('TIMEOUT: aborting connection')
    #    self.transport.abortConnection()

class TCPClientFactory(ClientFactory):

    def __init__(self, protocol):
        #print "ClientFactory build"
        self.protocol = protocol
        self.history = []
        self.cmd_list = []

    def buildProtocol(self, addr):
        #print "Connected to %s" % addr
        self.history.append('BUILDING PROTOCOL. ADDRESS = {}'.format(addr))
        p = self.protocol()
        p.factory = self
        p.cmdList = self.cmd_list
        self.cmd_list = []
        return p

    def clientConnectionFailed(self, connector, reason):
        """Clients call this when they are unable to initialize their connection."""
        self.history.append('CONNECTION FAILED: {}'.format(reason.getErrorMessage()))
        self.history.append('STOPPING REACTOR')
        reactor.stop()

    def clientConnectionLost(self, connector, reason):
        """Clients call this when their connections are lost."""
        #Note: could optionally use connector.connect() to try to rescue the session.
        self.history.append('CONNECTION LOST: {}'.format(reason.getErrorMessage()))
        self.history.append('STOPPING REACTOR')
        reactor.stop()





import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class FindPeaksByWindow(Operation):
    """
    Walk a 1d array and find its local maxima.
    A maximum is found if it is the highest point within windowsize of itself.
    An optional threshold for the peak intensity relative to the window-average
    can be used to filter out peaks due to noise.
    """

    def __init__(self):
        input_names = ['x','y','windowsize','threshold']
        output_names = ['pk_idx','x_pk','y_pk']
        super(FindPeaksByWindow,self).__init__(input_names, output_names)
        self.input_doc['x'] = '1d array of x values (domain- optional)'
        self.input_doc['y'] = '1d array of y values (amplitudes)'
        self.input_doc['windowsize'] = 'the window is this many points '\
            'in either direction of a given point'
        self.input_doc['threshold'] = 'threshold on Ipk/I(window) '\
            'for being counted as a peak: set to zero to deactivate'
        self.input_type['x'] = opmod.workflow_item
        self.input_type['y'] = opmod.workflow_item
        self.inputs['windowsize'] = 10
        self.inputs['threshold'] = 0 
        self.output_doc['pk_idx'] = 'q values of found peaks'
        self.output_doc['x_pk'] = 'x values of found peaks'
        self.output_doc['y_pk'] = 'y values of found peaks'

    def run(self):
        x = self.inputs['x']
        y = self.inputs['y']
        if x is None or y is None:
            return
        w = self.inputs['windowsize']
        thr = self.inputs['threshold']
        pk_idx = []
        for idx in range(w,len(y)-w-1):
            pkflag = False
            ywin = y[idx-w:idx+w+1]
            if np.argmax(ywin) == w:
                if thr:
                    pkflag = ywin[w]/np.mean(ywin) > thr
                else:
                    pkflag = True
            if pkflag:
                pk_idx.append(idx)
        self.outputs['pk_idx'] = np.array(pk_idx)
        self.outputs['x_pk'] = np.array([x[idx] for idx in pk_idx])
        self.outputs['y_pk'] = np.array([y[idx] for idx in pk_idx])
    




import numpy as np

from .. import Operation as opmod 
from ..Operation import Operation

class Zip(Operation):
    """
    Zip two 1d arrays together.
    """

    def __init__(self):
        input_names = ['x', 'y']
        output_names = ['x_y']
        super(Zip, self).__init__(input_names, output_names)
        self.input_doc['x'] = '1d array'
        self.input_doc['y'] = '1d array, same size as x'
        self.output_doc['x_y'] = 'n x 2 array containing x and y'
        # source & type
        self.input_type['x'] = opmod.workflow_item
        self.input_type['y'] = opmod.workflow_item

    def run(self):
        x = self.inputs['x']
        y = self.inputs['y']
        if x is None or y is None:
            return
        xy = np.array(zip(x, y))
        self.outputs['x_y'] = xy




import os.path

import numpy as np
import yaml
from sklearn import preprocessing
from sklearn import linear_model

from .... import Operation as opmod 
from ....Operation import Operation
from ..... import pawstools
from .....tools.saxs.saxs_classify import SaxsClassifier

class LoadSAXSClassifier(Operation):
    """
    Read files to load a set of classifiers to be used on 1-d saxs spectra. 
    """

    def __init__(self):
        input_names = ['file_path']
        output_names = ['saxs_classifier']
        super(LoadSAXSClassifier, self).__init__(input_names, output_names)
        self.input_doc['file_path'] = 'path to a pickle file '\
        'with data defining a set of scalers and classifiers '\
        'designed for 1-d (I(q) versus q) SAXS spectra. '\
        'If left blank, the default SAXS classifiers will be loaded.'
        self.output_doc['saxs_classifier'] = 'an object containing '\
            'scikit-learn classification models '\
            'designed for 1-d SAXS spectra.'
        self.inputs['file_path'] = os.path.join(pawstools.sourcedir,
        'core','tools','saxs','modeling_data','scalers_and_models.yml')

    def run(self):
        p = self.inputs['file_path']
        self.outputs['saxs_classifier'] = SaxsClassifier(p)





"""
Integrate an image to 2d,
using an existing PyFAI.AzimuthalIntegrator,
with a bunch of input parameters
for calling AzimuthalIntegrator.integrate1d().
"""

import numpy as np
import pyFAI

from ... import Operation as opmod 
from ...Operation import Operation

class ApplyIntegrator2d(Operation):
    """
    Input image data (ndarray), PyFAI.AzimuthalIntegrator, 
    mask, ROI mask, dark field image, flat field image,
    q-range, chi-range, number of points for integration bin centers,
    polz factor, choice of unit (string), 
    and choice of integration method (string).

    Refer to the PyFAI documentation at ..... 
    for parameter definitions and defaults.
    TODO: fill in web uri above. 
 
    Output arrays containing q, chi, and I(q,chi) 
    """
    def __init__(self):
        input_names = ['data','integrator','mask','ROI_mask','dark','flat',\
        'radial_range','azimuth_range','npt_rad','npt_azim','polarization_factor',\
        'normalization_factor','unit','method','integration_mode']
        output_names = ['q','chi','I_at_q_chi']
        super(ApplyIntegrator2d,self).__init__(input_names,output_names)
        self.input_doc['data'] = '2d array representing intensity for each pixel'
        self.input_doc['integrator'] = 'A PyFAI.AzimuthalIntegrator object'
        self.input_doc['mask'] = '2d array for image mask, same shape as image_data'
        self.input_doc['ROI_mask'] = '2d array for ROI mask, same shape as image_data'
        self.input_doc['dark'] = '2d array for dark field, same shape as image_data'
        self.input_doc['flat'] = '2d array for flat field, same shape as image_data'
        self.input_doc['radial_range'] = 'list with two values, lower and upper limits of q (scattering vector)'
        self.input_doc['azimuth_range'] = 'list with two values, lower and upper limits of chi (azimuthal angle)'
        self.input_doc['npt_rad'] = 'number of q-points to integrate, as evenly spaced bins between radial_range[0] and radial_range[1]'
        self.input_doc['npt_azim'] = 'number of q-points to integrate, as evenly spaced bins between radial_range[0] and radial_range[1]'
        self.input_doc['polarization_factor'] = 'polarization factor, if polarization correction is needed'
        self.input_doc['unit'] = 'choice of unit. See PyFAI documentation for options.'
        self.input_doc['method'] = 'choice of integration method. See PyFAI documentation for options.'
        self.input_doc['normalization_factor'] = 'normalization monitor value'
        self.input_doc['integration_mode'] = 'not yet implemented' 

        self.input_type['data'] = opmod.workflow_item
        self.input_type['integrator'] = opmod.workflow_item

        self.inputs['npt_rad'] = 1000
        self.inputs['npt_azim'] = 1000
        self.inputs['polarization_factor'] = 1.
        self.inputs['unit'] = 'q_A^-1'

        self.output_doc['q'] = 'Scattering vector magnitude q array in 1/Angstrom.'
        self.output_doc['chi'] = 'Azimuthal angle array.'
        self.output_doc['I_at_q_chi'] = '2d array of integrated intensity at q,chi.'

    def run(self):
        img = self.inputs['data']
        intgtr = self.inputs['integrator']
        if img is None or intgtr is None:
            return
        m = None
        if self.inputs['mask'] is not None and self.inputs['ROI_mask'] is not None: 
            m = self.inputs['mask'] | self.inputs['ROI_mask']
        elif self.inputs['mask'] is not None:
            m = self.inputs['mask']
        elif self.inputs['ROI_mask'] is not None:
            m = self.inputs['ROI_mask']
        self.inputs['mask'] = m
        #if self.inputs['ROI_mask']: self.inputs['mask'] = self.inputs['mask'] | self.inputs['ROI_mask']
        kwargexcludemask = [k for k in self.inputs.keys() if self.inputs[k] is None]
        kwargexcludemask = kwargexcludemask + ['ROI_mask','integrator','integration_mode']
        kwargs = {key:val for key,val in self.inputs.items() if key not in kwargexcludemask}

        I_at_q_chi,q,chi = intgtr.integrate2d(**kwargs)
        #integ_result = intgtr.integrate2d(**kwargs)
        #q = integ_result.radial
        #chi = integ_result.chi
        #I_at_q_chi = integ_result.intensity
        self.outputs['q'] = q
        self.outputs['chi'] = chi
        self.outputs['I_at_q_chi'] = I_at_q_chi




from os.path import splitext, split
from collections import OrderedDict

from ... import Operation as opmod 
from ...Operation import Operation

class ReadHeader_SSRL15(Operation):
    """
    Read a .txt header from beamline 1-5 at SSRL into a dict.
    """

    def __init__(self):
        input_names = ['file_path']
        output_names = ['header_dict', 'filename']
        super(ReadHeader_SSRL15, self).__init__(input_names, output_names)
        self.input_doc['file_path'] = 'path to a .txt header file produced by beamline 1-5 at SSRL.'
        self.output_doc['header_dict'] = 'the header file as a python dictionary'
        self.output_doc['filename'] = 'filename with path and extension stripped'

    def run(self):
        file_path = self.inputs['file_path']
        if file_path is None:
            return
        filename = split(file_path)[1]
        filename_noext = splitext(filename)[0]
        self.outputs['filename'] = filename_noext 
        d = OrderedDict()
        for l in open(file_path,'r').readlines():
            try:
                if not l.strip() == '' and not l.strip()[0] == '#':
                    kvs = l.split(',')
                    # special case for the string headers on line 1
                    if not kvs[0].find('User') == -1:
                        u_str = kvs[0].split('User:')[1].strip()
                        t_str = kvs[1].split('time:')[1].strip()
                        d['User'] = u_str
                        d['time'] = t_str
                    # and filter out the redundant temperature line
                    elif not (len(kvs)==1 and kvs[0].strip()[-1]=='C'):
                        for kv in kvs:
                            kv_arr = kv.split('=')
                            d[kv_arr[0].strip()] = float(kv_arr[1].strip())
                self.outputs['header_dict'] = d
            except Exception as e:
                d['ERROR'] = e.message
                self.outputs['header_dict'] = d
                raise e




import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class ArrayLog(Operation):
    """
    Take the base-10 logarithm of any array. 
    Any elements with non-positive values are removed. 
    """

    def __init__(self):
        input_names = ['x']
        output_names = ['logx']
        super(ArrayLog, self).__init__(input_names, output_names)
        self.input_doc['x'] = 'any array'
        self.output_doc['logx'] = 'array of log(x), same size as x, non-positive values replaced with np.nan'
        self.input_type['x'] = opmod.workflow_item

    def run(self):
        x = self.inputs['x']
        if x is None:
            return 
        # good_vals = elements for which both x and y have defined logarithm
        idx_ok = ((x > 0) & (~np.isnan(x)))
        logx = np.zeros(x.shape)
        logx[idx_ok] = np.log10(x[idx_ok])
        logx[np.invert(idx_ok)] = np.nan
        self.outputs['logx'] = logx 





import os

import fabio

from ... import Operation as opmod 
from ...Operation import Operation

class FabIOOpen(Operation):
    """
    Takes a filesystem path and calls fabIO to load it. 
    """

    def __init__(self):
        input_names = ['path']
        output_names = ['image_data','FabioImage','dir_path','filename']
        super(FabIOOpen,self).__init__(input_names,output_names) 
        self.input_doc['path'] = 'string representing the path to a .tif image'
        self.output_doc['image_data'] = '2D array representing pixel values taken from the input file'
        self.output_doc['FabioImage'] = 'The object generated by fabio.open()'
        self.output_doc['dir_path'] = 'Path to the directory the image came from'
        self.output_doc['filename'] = 'The image filename, no path, no extension'
        
    def run(self):
        """
        Call on fabIO to extract image data
        """
        p = self.inputs['path']
        if p is None:
            return
        dir_path = os.path.split(p)[0]
        file_nopath = os.path.split(p)[1]
        file_noext = os.path.splitext(file_nopath)[0]
        self.outputs['dir_path'] = dir_path 
        self.outputs['filename'] = file_noext 
        self.outputs['image_data'] = fabio.open(p).data




import numpy as np
import pyFAI

from ... import Operation as opmod 
from ...Operation import Operation

class WXDToPONI(Operation):
    """
    Convert WXDIFF .calib output 
    to a dict of PyFAI PONI parameters,
    by first converting from WXDIFF to Fit2D,
    then using a pyFAI.AzimuthalIntegrator 
    to convert from Fit2D to PONI format.
    
    The conversion from WXDiff parameters to Fit2D parameters 
    was originally contributed to paws by Fang Ren.

    Input .calib file from WXDIFF automated calibration,
    input pixel size and polarization factor,
    output dict of pyFAI PONI calibration parameters.
    """
    
    def __init__(self):
        input_names = ['wxd_file','pixel_size_um','fpolz']
        output_names = ['poni_dict']
        super(WXDToPONI,self).__init__(input_names,output_names)
        self.input_doc['wxd_file'] = '.calib file produced by WXDIFF calibration'
        self.input_doc['pixel_size_um'] = 'pixel size in microns'
        self.input_doc['fpolz'] = 'polarization factor'
        self.inputs['fpolz'] = float(1.)
        self.output_doc['poni_dict'] = 'Dict of pyFAI calibration parameters, as found in a .poni file'

    def run(self):
        fpath = self.inputs['wxd_file']
        pxsz_um = self.inputs['pixel_size_um']
        if fpath is None or pxsz_um is None:
            return
        pxsz_m = pxsz_um*1E-6
        fpolz = self.inputs['fpolz']
        for line in open(fpath,'r'):
            kv = line.strip().split('=')
            if kv[0] == 'detect_dist':
                d_px = float(kv[1])         # WXDIFF direct detector distance, 
                                            # from sample to where beam axis intersects detector plane, in pixels.
                                            # Fit2D uses this input in mm to set its spatial scale,
                                            # so it has to be converted to distance units (mm for Fit2D). 
                                            # Because this scales the spatial representation for Fit2D,
                                            # the other inputs can still be given in pixel units,
                                            # without converting between pixel sizes.
                d_m = d_px*pxsz_m 
                d_mm = d_m*1E3 
            if kv[0] == 'bcenter_x':
                bcx_px = float(kv[1])       # WXDIFF x coord relative to 'bottom left' corner of detector
                                            # where beam axis intersects detector plane, in pixels 
                                            # Fit2D centerX is also in pixels 
            if kv[0] == 'bcenter_y':
                bcy_px = float(kv[1])       # WXDIFF y coord relative to 'bottom left' corner of detector
                                            # where beam axis intersects detector plane, in pixels 
                                            # Fit2D centerY is also in pixels
            if kv[0] == 'detect_tilt_alpha':    
                rot_rad = float(kv[1])      # WXDIFF alpha is tilt plane rotation in rad 
                                            # Fit2D tiltPlanRotation  = 360-rot_rad*180/pi 
            if kv[0] == 'detect_tilt_delta':
                tilt_rad = float(kv[1])     # WXDIFF delta is the tilt in rad
            if kv[0] == 'wavelenght':
                wl_A = float(kv[1])         # WXDIFF 'wavelenght' is in Angstroms
        # get wavelength in m 
        wl_m = wl_A*1E-10
        # Mapping between Fit2D and WXDIFF originally contributed by Fang Ren. 
        # TODO: Verify this mapping
        rot_deg = rot_rad*float(180)/np.pi
        rot_fit2d = float(360)-rot_deg
        tilt_deg = tilt_rad*float(180)/np.pi
        # use a pyFAI.AzimuthalIntegrator() to do the conversion
        p = pyFAI.AzimuthalIntegrator(wavelength = wl_m) 
        p.setFit2D(d_mm,bcx_px,bcy_px,tilt_deg,rot_fit2d,pxsz_um,pxsz_um)
        poni_dict = p.getPyFAI()
        poni_dict['fpolz'] = fpolz
        self.outputs['poni_dict'] = poni_dict 




from __future__ import print_function
from collections import OrderedDict
from functools import partial
import copy

from PySide import QtCore

from .QWfWorker import QWfWorker
from .QWorkflow import QWorkflow
from ..core import pawstools
from ..core.workflow.WfManager import WfManager
from . import qttools

class QWfManager(WfManager,QtCore.QObject):
    """
    A Qt Signal-slot manager for paws Workflows.
    """
    # this signal should emit the name
    # of the new workflow 
    wfAdded = QtCore.Signal(str)

    # this signal should emit the name
    # of the workflow that finished.
    wfFinished = QtCore.Signal(str)
    wfStopped = QtCore.Signal(str)

    # a signal for carrying textual output
    # e.g. for message board logging
    emitMessage = QtCore.Signal(str)

    @QtCore.Slot(str)
    def relayMessage(self,msg):
        self.emitMessage.emit(msg)

    @QtCore.Slot(str)
    def launchWorkflow(self,wfname):
        self.run_wf(wfname)

    def __init__(self,qapp):
        super(QWfManager,self).__init__()
        self.app = qapp 
        self.wf_running = OrderedDict()

    def add_wf(self,wfname):
        """
        Add a QWorkflow to self.workflows, with key specified by wfname.
        If wfname is not unique (i.e. a workflow with that name already exists),
        this method will overwrite the existing workflow with a new one.
        """
        wf = QWorkflow()
        if not wf.is_tag_valid(wfname): 
            raise pawstools.WfNameError(wf.tag_error_message(wfname))
        wf.message_callback = wf.emitMessage.emit
        wf.emitMessage.connect( self.relayMessage )
        #wf.data_callback = wf.emitData.emit
        #wf.emitData.connect( wf.updateItem )
        wf.data_callback = wf.updateItem
        wf.wfFinished.connect( partial(self.stop_wf,wfname) )
        self.workflows[wfname] = wf
        self.wf_running[wfname] = False
        self.wfAdded.emit(wfname)

    def stop_wf(self,wfname):
        self.wf_running[wfname] = False
        self.wfStopped.emit(wfname)

    def run_wf(self,wfname,pool=None):
        wf = self.workflows[wfname]
        stk,diag = wf.execution_stack()
        self.logmethod('preparing workflow {} for execution'.format(wfname))
        self.prepare_wf(wf,stk)
        if pool is None:
            #wf.message_callback = self.logmethod
            #wf.data_callback = None
            self.wf_running[wfname] = True
            wf.execute()
            self.logmethod('execution finished')
        else:
            # Copy the workflow so it can be moved off the main thread.
            # Connect some signals so that the local copy gets updated
            # as the remote copy gets executed. 
            # NOTE: This implementation assumes the Workflows are QWorkflows.
            wf_clone = wf.clone_wf()
            wf_clone.message_callback = wf_clone.emitMessage.emit
            wf_clone.emitMessage.connect(self.relayMessage)
            wf_clone.data_callback = wf_clone.emitData.emit
            wf_clone.emitData.connect(wf.updateItem)
            wf_clone.wfFinished.connect( partial(self.stop_wf,wfname) )
            # construct a QRunnable around a copy of the Workflow 
            qr = qttools.RunnableExecutor( wf_clone ) 
            # Call QThreadPool.start(QRunnable) - executes runnable.run()
            self.wf_running[wfname] = True
            pool.start(qr)






import pypif.obj as pifobj

from ... import Operation as opmod 
from ...Operation import Operation

class EmptyPif(Operation):
    """
    Make and empty pypif.obj.ChemicalSystem object.
    """

    def __init__(self):
        input_names = ['uid','date_time','t_utc']
        output_names = ['pif']
        super(EmptyPif,self).__init__(input_names,output_names)
        self.input_doc['uid'] = 'text string to use as pif record uid'
        self.input_doc['date_time'] = 'string date/time for pif record tags'
        self.input_doc['t_utc'] = 'time of record creation in utc'
        self.output_doc['pif'] = 'an empty pif object'
        self.input_type['date_time'] = opmod.workflow_item
        self.input_type['t_utc'] = opmod.workflow_item

    def run(self):
        uid = self.inputs['uid']
        t_str = self.inputs['date_time']
        t_utc = self.inputs['t_utc']
        if uid is None:
            return
        main_sys = pifobj.ChemicalSystem()
        main_sys.uid = uid
        main_sys.tags = ['experiment id: '+uid]
        if t_str is not None:
            main_sys.tags.append('date: '+t_str)
        if t_utc is not None:
            main_sys.tags.append('utc: '+str(int(t_utc)))
        self.outputs['pif'] = main_sys

    def saxs_to_pif_properties(self,q_I,T_C):
        #props = []
        #for i in range(len(q)):
        pq = pifobj.Property()
        n_qpoints = q_I.shape[0]
        ### property: scattered intensity
        pI = pifobj.Property()
        pI.name = 'SAXS intensity'
        pI.scalars = [pifobj.Scalar(q_I[i,1]) for i in range(n_qpoints)]
        pI.units = 'counts'
        pI.conditions = []
        pI.conditions.append( pifobj.Value('SAXS scattering vector', 
                            [pifobj.Scalar(q_I[i,0]) for i in range(n_qpoints)],
                            None,None,'1/Angstrom') )
        pI.conditions.append(pifobj.Value('temperature',[pifobj.Scalar(T_C)],None,None,'degrees Celsius'))
        return [pq,pI] 
        
#    def make_piftemperature(self,t):
#        v = pifobj.Value()
#        v.name = 'temperature'
#        tscl = pifobj.Scalar()
#        tscl.value = str(t)
#        v.scalars = [tscl]
#        v.units = 'degrees Celsius'
#        return v
#
#    def make_pifvector(self,v):
#        pifv = []
#        for n in v:
#            s = pifobj.Scalar()
#            s.value = str(n)
#            pifv.append(s)
#        return pifv
#
#    def pifscalar(self,scl,errlo,errhi):
#        s = pifobj.Scalar()
#        s.value = str(scl) 
#        s.minimum = str(scl) 
#        s.maximum = str(scl) 
#        s.inclusiveMinimum = True
#        s.inclusiveMaximum = True
#        s.uncertainty = '+{},-{}'.format(errlo,errhi)
#        s.approximate = True
#        return s






from fabio import edfimage
from fabio import tifimage

from ... import Operation as opmod 
from ...Operation import Operation
import os

class FabIOWrite(Operation):
    """
    Use FabIO to write out an image, given image data,
    directory path, filename, file tag, extension, 
    an image header (dict), and a flag for whether or not to overwrite. 

    Outputs the full file path where the image was written,
    which should be dir_path+filename+filetag+ext.
    """

    def __init__(self):
        input_names = ['image_data','header','dir_path','filename','filetag','ext','overwrite']
        output_names = ['file_path']
        super(FabIOWrite,self).__init__(input_names,output_names)
        self.input_doc['image_data'] = 'image/array data to be saved'
        self.input_doc['header'] = 'dict-like image metadata header'
        self.input_doc['dir_path'] = 'parent directory path'
        self.input_doc['filename'] = 'base filename to be saved as'
        self.input_doc['filetag'] = 'string tag to be appended to base filename'
        self.input_doc['ext'] = 'file extension (overwrites base filename extension)'
        self.input_doc['overwrite'] = 'allow overwrite of already existing files'
        self.output_doc['file_path'] = 'path to the file that will be written: dir_path+filename+filetag+ext'

        self.input_type['image_data'] = opmod.workflow_item
        self.input_type['header'] = opmod.workflow_item
        self.input_type['filename'] = opmod.workflow_item

        self.inputs['filetag'] = ''
        self.inputs['ext'] = '.tif'
        self.inputs['overwrite'] = False

    def run(self):
        """
        Call on fabIO to extract image data
        """
        img = self.inputs['image_data']
        p = self.inputs['dir_path']
        fnm = self.inputs['filename'] 
        if p is None or fnm is None or img is None:
            return
        ext = self.inputs['ext'].lower()
        tg = self.inputs['filetag']
        hdr = self.inputs['header']
        outfile = fnm + tg + ext
        filepath = os.path.join(self.inputs['dir_path'],outfile)
        self.outputs['file_path'] = filepath
        if os.path.isfile(filepath) and not self.inputs['overwrite']:
            raise IOError('File already exists.')

        cls = None

        if ext == '.edf': cls = edfimage.edfimage
        elif ext == '.tif': cls = tifimage.tifimage
        # TODO: add JPEG support

        if not cls: raise ValueError('Extension not supported.')

        cls(img,header=hdr).write(filepath)





from PySide import QtCore, QtGui
import pypif.obj as pifobj

from text_widgets import display_text, unit_indent

# TODO: Make an (editable?) PIF Widget

class PifWidget(QtGui.QTextEdit):
    
    def __init__(self,itm):
        t = self.print_pif(itm,unit_indent)
        super(PifWidget,self).__init__()
        self.setText(t) 

    def print_pif(self,itm,indent):
        t = '(pypif.obj.System)'
        if itm.uid: 
            #string
            t += '<br>' + indent + 'uid: {}'.format(display_text(itm.uid,indent+unit_indent))
        if isinstance(itm,pifobj.ChemicalSystem):
            if itm.chemical_formula: 
                #string
                t += '<br>' + indent + 'chemical_formula: {}'.format(
                display_text(itm.chemical_formula,indent+unit_indent))
            if itm.composition is not None:
                #array of pypif.obj.Composition
                t += '<br>' + indent + 'composition: (array)'
                for i,comp in zip(range(len(itm.composition)),itm.composition):
                    t += '<br>' + indent + unit_indent + '{}: {}'.format(
                    i,self.print_comp(comp,indent+unit_indent+unit_indent))
        if itm.names is not None: 
            #array of string
            t += '<br>' + indent + 'names: (array)'
            for i,nm in zip(range(len(itm.names)),itm.names):
                t += '<br>' + indent + unit_indent + '{}: {}'.format(
                i, display_text(nm,indent+unit_indent+unit_indent))
        if itm.quantity: 
            #pypif.obj.Quantity
            t += '<br>' + indent + 'quantity: {}'.format(self.print_qty(itm.quantity,indent+unit_indent))
        if itm.source: 
            #pypif.obj.Source
            t += '<br>' + indent + 'source: {}'.format(self.print_pifsrc(itm.source,indent+unit_indent))
        if itm.preparation is not None: 
            #array of pypif.obj.ProcessStep
            t += '<br>' + indent + 'preparation: (array)'
            for i,procstep in zip(range(len(itm.preparation)),itm.preparation):
                t += '<br>' + indent + unit_indent + '{}: {}'.format(
                i, self.print_procstep(procstep,indent+unit_indent+unit_indent))
        if itm.properties is not None: 
            #array of pypif.obj.Property
            t += '<br>' + indent + 'properties: (array)'
            for i,prop in zip(range(len(itm.properties)),itm.properties):
                t += '<br>' + indent + unit_indent + '{}: {}'.format(
                i, self.print_prop(prop,indent+unit_indent+unit_indent))
        if itm.tags is not None: 
            #array of string
            t += '<br>' + indent + 'tags: (array)'
            for i,tg in zip(range(len(itm.tags)),itm.tags):
                t += '<br>' + indent + unit_indent + '{}: {}'.format(
                i, display_text(tg,indent+unit_indent+unit_indent))
        if itm.ids is not None: 
            #array of pypif.obj.Id
            t += '<br>' + indent + 'ids: (array)'
            for i,id_ in zip(range(len(itm.ids)),itm.ids):
                t += '<br>' + indent + unit_indent + '{}: {}'.format(
                i, self.print_id(id_,indent+unit_indent+unit_indent))
        if itm.sub_systems is not None: 
            #array of pypif.obj.System
            t += '<br>' + indent + 'sub_systems: (array)'
            for i,sys in zip(range(len(itm.sub_systems)),itm.sub_systems):
                t += '<br>' + indent + unit_indent + '{}: {}'.format(
                i, self.print_pif(sys,indent+unit_indent+unit_indent))
        return t    

    def print_id(self,id_,indent):
        t = '(pypif.obj.Id)'
        return t
    
    def print_comp(self,itm,indent):    
        t = '(pypif.obj.Composition)'
        return t
    
    def print_qty(self,itm,indent):    
        t = '(pypif.obj.Quantity)'
        return t

    def print_pifsrc(self,itm,indent):    
        t = '(pypif.obj.Source)'
        return t

    def print_procstep(self,itm,indent):
        t = '(pypif.obj.ProcessStep)'
        return t

    def print_prop(self,itm,indent):
        t = '(pypif.obj.Property)'
        t += '<br>' + indent + 'name: {}'.format(display_text(itm.name,indent+unit_indent))
        if itm.conditions is not None:
            t += '<br>' + indent + 'conditions: (array)' 
            for i,val in zip(range(len(itm.conditions)),itm.conditions):
                t += '<br>' + indent + unit_indent + '{}: {}'.format(
                i, self.print_value(val,indent+unit_indent+unit_indent))
        if itm.scalars is not None:
            t += '<br>' + indent + 'scalars: (array)' 
            for i,scl in zip(range(len(itm.scalars)),itm.scalars):
                t += '<br>' + indent + unit_indent + '{}: {}'.format(
                i, self.print_scalar(scl,indent+unit_indent+unit_indent))
        return t

    def print_value(self,itm,indent):
        t = '(pypif.obj.Value)'
        t += '<br>' + indent + 'name: {}'.format(display_text(itm.name,indent+unit_indent))
        if itm.scalars is not None:
            t += '<br>' + indent + 'scalars: (array)'
            for i,scl in zip(range(len(itm.scalars)),itm.scalars):
                t += '<br>' + indent + unit_indent + '{}: {}'.format(
                i, self.print_scalar(scl,indent+unit_indent+unit_indent))
        elif itm.vectors is not None:
            t += '<br>' + indent + 'vectors: (array)'
            for i,vec in zip(range(len(itm.vectors)),itm.vectors):
                t += '<br>' + indent + unit_indent + '{}: {}'.format(
                i, self.print_vector(vec,indent+unit_indent+unit_indent))
        elif itm.matrices is not None:
            t += '<br>' + indent + 'matrices: (array)'
            for i,mat in zip(range(len(itm.matrices)),itm.matrices):
                t += '<br>' + indent + unit_indent + '{}: {}'.format(
                i, self.print_matrix(mat,indent+unit_indent+unit_indent))
        return t

    def print_scalar(self,itm,indent):
        t = '(pypif.obj.Scalar)'
        t += '<br>' + indent + 'value: {}'.format(itm.value) 
        return t

    def print_vector(self,itm,indent):
        t = '(array)'
        for i,scl in zip(range(len(itm)),itm):
            t += '<br>' + indent + unit_indent + '{}: {}'.format(
            i, self.print_scalar(scl,indent+unit_indent+unit_indent))
        return t

    def print_matrix(self,itm,indent):
        t = '(array)'
        for i,vec in zip(range(len(itm)),itm):
            t += '<br>' + indent + unit_indent + '{}: {}'.format(
            i, self.print_vector(vec,indent+unit_indent+unit_indent))
        return t




from __future__ import print_function
from collections import OrderedDict
import copy
from functools import partial
import traceback
import os

from ..models.TreeModel import TreeModel
from ..operations import Operation as opmod
from ..operations.Operation import Operation
from ..operations import optools

class Workflow(TreeModel):
    """
    Tree structure for a Workflow built from paws Operations.
    """

    def __init__(self):
        flag_dict = OrderedDict()
        flag_dict['select'] = False
        flag_dict['enable'] = True
        super(Workflow,self).__init__(flag_dict)
        self.inputs = OrderedDict()
        self.outputs = OrderedDict()
        self.message_callback = print
        self.data_callback = None

    def __getitem__(self,key):
        optags = self.keys()
        if key in optags:
            return self.get_data_from_uri(key) 
        else:
            raise KeyError('[{}] {}.__getitem__ only recognizes keys {}'
            .format(__name__,type(self).__name__,optags))
    def __setitem__(self,key,data):
        optags = self.keys() 
        # TODO: ensure that data is an Operation?
        if key in optags:
            self.set_item(key,data)
        else:
            raise KeyError('[{}] {}.__setitem__ only recognizes keys {}'
            .format(__name__,type(self).__name__,optags))

    def keys(self):
        return self.list_op_tags() 

    def set_op_item(self,op_tag,item_uri,item_data):
        full_uri = op_tag+'.'+item_uri
        self.set_item(full_uri,item_data)

    def clone_wf(self):
        """
        Produce a Workflow that is a copy of this Workflow.
        """
        new_wf = self.clone() 
        #new_wf = copy.copy(self)
        new_wf.inputs = copy.deepcopy(self.inputs)
        new_wf.outputs = copy.deepcopy(self.outputs)
        # NOTE: is it ok if I don't copy.copy the callbacks? 
        new_wf.message_callback = self.message_callback
        new_wf.data_callback = self.data_callback
        for op_tag in self.list_op_tags():
            op = self.get_data_from_uri(op_tag)
            new_wf.add_op(op_tag,op.clone_op())
        return new_wf

    @classmethod
    def clone(cls):
        return cls()

    def add_op(self,op_tag,op):
        op.message_callback = self.message_callback
        op.data_callback = partial( self.set_op_item,op_tag )
        self.set_item(op_tag,op)

    def build_tree(self,x):
        """
        Reimplemented TreeModel.build_tree() 
        so that TreeItems are built from Operations.
        """
        if isinstance(x,Operation):
            d = OrderedDict()
            d['inputs'] = self.build_tree(x.inputs)
            d['outputs'] = self.build_tree(x.outputs)
            return d
        else:
            return super(Workflow,self).build_tree(x) 

    def op_dict(self):
        optags = self.list_op_tags() 
        return OrderedDict(zip(optags,[self.get_data_from_uri(nm) for nm in optags]))

    def list_op_tags(self):
        return self.root_tags()

    def n_ops(self):
        return self.n_children()

    def connect_wf_input(self,wf_input_name,op_input_uris):
        self.inputs[wf_input_name] = op_input_uris

    def connect_wf_output(self,wf_output_name,op_output_uris):
        self.outputs[wf_output_name] = op_output_uris

    def break_wf_input(self,wf_input_name):
        self.inputs.pop(wf_input_name)
    
    def break_wf_output(self,wf_output_name):
        self.outputs.pop(wf_output_name)

    def wf_outputs_dict(self):
        d = OrderedDict()
        for wfoutnm in self.outputs.keys():
            # if the uri referred to by this output exists, save it
            if self.contains_uri(self.outputs[wfoutnm]):
                d[wfoutnm] = self.get_data_from_uri(self.outputs[wfoutnm])
        return d

    def get_wf_output(wf_output_name):
        """
        Fetch and return the Operation output(s)
        indicated by self.outputs[wf_output_name].
        """
        r = self.outputs[wf_output_name]
        if isinstance(r,list):
            return [self.get_data_from_uri(q) for q in r]
        else:
            return self.get_data_from_uri(r)

    def set_wf_input(self,wf_input_name,val):
        """
        Take the Operation input(s) 
        indicated by self.inputs[wf_input_name],
        and set them to the input value val. 
        """
        urilist = self.inputs[wf_input_name]
        if not isinstance(urilist,list):
            urilist = [urilist]
        for uri in urilist:
            p = uri.split('.')
            il = self.get_data_from_uri(p[0]).input_locator[p[2]]
            if il.tp in [opmod.basic_type,opmod.runtime_type]:
                self.set_item(uri,val)
            else:
                il.val = val
                #msg = 'Attempted to set input {}, '.format(uri)\
                #'and found that input was supposed to be a {}. '.format(opmod.input_types[il.tp])\
                #'Only "auto" type inputs should be loaded directly.'
                #raise TypeError(msg)

    def execute(self):
        stk,diag = self.execution_stack()
        bad_diag_keys = [k for k in diag.keys() if diag[k]]
        for k in bad_diag_keys:
            self.message_callback('WARNING- operation not ready: {}'.format(diag[k]))
        self.message_callback('workflow queue:'+os.linesep+self.print_stack(stk))
        for lst in stk:
            self.message_callback('running: {}'.format(lst))
            for op_tag in lst: 
                op = self.get_data_from_uri(op_tag) 
                for inpnm,il in op.input_locator.items():
                    if il.tp == opmod.workflow_item:
                        #il.data = self.locate_input(il)
                        #op.inputs[inpnm] = il.data
                        #op.inputs[inpnm] = self.locate_input(il)
                        self.set_op_item(op_tag,'inputs.'+inpnm,self.locate_input(il))
                op.run() 
                for outnm,outdata in op.outputs.items():
                    self.set_op_item(op_tag,'outputs.'+outnm,outdata)

    def locate_input(self,il):
        if isinstance(il.val,list):
            return [self.get_data_from_uri(v) for v in il.val]
        else:
            return self.get_data_from_uri(il.val)

    def set_op_enabled(self,opname,flag=True):
        op_item = self.get_from_uri(opname)
        op_item.flags['enable'] = flag

    def is_op_enabled(self,opname):
        op_item = self.get_from_uri(opname)
        return op_item.flags['enable']

    def op_enable_flags(self):
        dct = OrderedDict()
        for opnm in self.list_op_tags():
            dct[opnm] = self.get_from_uri(opnm).flags['enable']
        return dct

    def execution_stack(self):
        """
        Build a stack (list) of lists of Operation uris,
        such that each list indicates a set of Operations
        whose dependencies are satisfied by the Operations above them.
        """
        stk = []
        valid_wf_inputs = [] 
        diagnostics = {}
        continue_flag = True
        while not self.stack_size(stk) == self.n_ops() and continue_flag:
            ops_rdy = []
            ops_not_rdy = []
            for op_tag in self.list_op_tags():
                if not self.is_op_enabled(op_tag):
                    op_rdy = False
                    op_diag = {op_tag:'Operation is disabled'}
                elif not self.stack_contains(op_tag,stk):
                    op = self.get_data_from_uri(op_tag)
                    op_rdy,op_diag = self.is_op_ready(op_tag,self,valid_wf_inputs)
                    diagnostics.update(op_diag)
                    if op_rdy:
                        ops_rdy.append(op_tag)
                    else:
                        ops_not_rdy.append(op_tag)
            if any(ops_rdy):
                stk.append(ops_rdy)
                for op_tag in ops_rdy:
                    op = self.get_data_from_uri(op_tag)
                    valid_wf_inputs += self.get_valid_wf_inputs(op_tag,op)
            else:
                continue_flag = False
        return stk,diagnostics

    def wf_setup_dict(self):
        wf_dict = OrderedDict() 
        for opname in self.list_op_tags():
            op = self.get_data_from_uri(opname)
            wf_dict[opname] = op.setup_dict()
        wf_dict['WORKFLOW_INPUTS'] = self.inputs
        wf_dict['WORKFLOW_OUTPUTS'] = self.outputs
        wf_dict['OP_ENABLE_FLAGS'] = self.op_enable_flags()
        return wf_dict

    def build_op_from_dict(self,op_setup,op_manager):
        op_uri = op_setup['op_module']
        op_manager.set_op_enabled(op_uri)
        op = op_manager.get_data_from_uri(op_uri)()
        op.load_defaults()
        il_setup_dict = op_setup['inputs']
        for nm in op.inputs.keys():
            if nm in il_setup_dict.keys():
                tp = il_setup_dict[nm]['tp']
                val = il_setup_dict[nm]['val']
                op.input_locator[nm] = opmod.InputLocator(tp,val) 
        return op

    @staticmethod
    def stack_contains(itm,stk):
        for lst in stk:
            if itm in lst:
                return True
            for lst_itm in lst:
                if isinstance(lst_itm,list):
                    if stack_contains(itm,lst_itm):
                        return True
        return False

    @staticmethod
    def stack_size(stk):
        sz = 0
        for lst in stk:
            for lst_itm in lst:
                if isinstance(lst_itm,list):
                    sz += stack_size(lst_itm)
                else:
                    sz += 1
        return sz

    @staticmethod
    def is_op_ready(op_tag,wf,valid_wf_inputs):
        op = wf.get_data_from_uri(op_tag)
        inputs_rdy = []
        diagnostics = {} 
        for name,il in op.input_locator.items():
            msg = ''
            if il.tp == opmod.workflow_item:
                inp_rdy = False
                if isinstance(il.val,list):
                    if all([v in valid_wf_inputs for v in il.val]):
                        inp_rdy = True 
                else:
                    if il.val in valid_wf_inputs: 
                        inp_rdy = True 
                if not inp_rdy:
                    msg = str('Operation input {} (={}) '.format(name,il.val)
                    + 'not satisfied by valid inputs list: {}'.format(valid_wf_inputs))
            else:
                inp_rdy = True
            inputs_rdy.append(inp_rdy)
            diagnostics[op_tag+'.inputs.'+name] = msg
        if all(inputs_rdy):
            op_rdy = True
        else:
            op_rdy = False
        return op_rdy,diagnostics 

    @staticmethod
    def get_valid_wf_inputs(op_tag,op):
        """
        Return the TreeModel uris of the op and its inputs/outputs 
        that are eligible as downstream inputs in the workflow.
        """
        # valid_wf_inputs should be the operation, its input and output dicts, and their respective entries
        valid_wf_inputs = [op_tag,op_tag+'.inputs',op_tag+'.outputs']
        valid_wf_inputs += [op_tag+'.outputs.'+k for k in op.outputs.keys()]
        valid_wf_inputs += [op_tag+'.inputs.'+k for k in op.inputs.keys()]
        return valid_wf_inputs

    @staticmethod
    def print_stack(stk):
        stktxt = ''
        opt_newline = '\n'
        for i,lst in zip(range(len(stk)),stk):
            if i == len(stk)-1:
                opt_newline = ''
            if len(lst) > 1:
                if isinstance(lst[1],list):
                    substk = lst[1]
                    stktxt += ('[\'{}\':\n{}\n]'+opt_newline).format(lst[0],print_stack(lst[1]))
                else:
                    stktxt += ('{}'+opt_newline).format(lst)
            else:
                stktxt += ('{}'+opt_newline).format(lst)
        return stktxt





from __future__ import print_function
import abc
import re
from collections import OrderedDict
import copy

# Enumeration of valid types for workflow and plugin inputs
no_input = 0            # ensure the input is set to None 

basic_type = 1          # input is set by hand, 
                        # and is of a basic type
                        # that is easy to serialize 

workflow_item = 2       # the address (TreeModel uri)
                        # of an item in the workflow 

entire_workflow = 3     # the name of a Workflow
                        # in the current WfManager

plugin_item = 4         # the address (TreeModel uri)
                        # of an item in the PluginManager

runtime_type = 5        # input is generated and set during execution,
                        # and serialization should not be attempted


valid_types = [no_input,basic_type,workflow_item,entire_workflow,plugin_item,runtime_type]
input_types = ['none','basic','workflow item','entire workflow','plugin item','runtime']

class InputLocator(object):
    """
    Objects of this class are used as containers for inputs to an Operation.
    They contain the information needed to find the relevant input data.
    """
    def __init__(self,tp=no_input,val=None):
        self.tp = tp
        self.val = val 
        #self.data = None 

class Operation(object):
    """
    Class template for implementing paws operations.
    """

    def __init__(self,input_names,output_names):
        """
        The input_names and output_names (lists of strings)
        are used to specify names for the parameters 
        that will be used to perform the operation.
        These lists are used as keys to build dicts
        Operation.inputs and Operation.outputs.
        """
        self.inputs = OrderedDict()
        self.input_locator = OrderedDict() 
        self.outputs = OrderedDict() 
        self.input_doc = OrderedDict() 
        self.input_type = OrderedDict() 
        self.output_doc = OrderedDict() 
        # For each of the i/o names, assign to None 
        for name in input_names: 
            self.input_type[name] = basic_type 
            self.input_locator[name] = None 
            self.inputs[name] = None
            self.input_doc[name] = None
        for name in output_names: 
            self.outputs[name] = None
            self.output_doc[name] = None
        self.message_callback = print
        self.data_callback = None 
    
    def __getitem__(self,key):
        if key == 'inputs':
            return self.inputs
        elif key == 'outputs':
            return self.outputs
        else:
            raise KeyError('[{}] Operation only recognizes keys {}'
            .format(__name__,self.keys()))
    def __setitem__(self,key,data):
        if key == 'inputs':
            self.inputs = data
        elif key == 'outputs':
            self.outputs = data
        else:
            raise KeyError('[{}] Operation only recognizes keys {}'
            .format(__name__,self.keys()))
    def keys(self):
        return ['inputs','outputs']

    def load_defaults(self):
        """
        Set default types and values into the Operation.input_locators.
        """
        for name in self.inputs.keys():
            tp = self.input_type[name]
            val = None
            if self.inputs[name] is not None:
                val = self.inputs[name]
            self.input_locator[name] = InputLocator(tp,val)

    def run(self):
        """
        Operation.run() should use the Operation.inputs
        and set values for all of the items in Operation.outputs.
        """
        pass

    @classmethod
    def clone(cls):
        return cls()

    def clone_op(self):
        """
        Clone the Operation.
        This should be called after all inputs have been loaded,
        with the exception of workflow items,
        e.g. after calling WfManager.prepare_wf().
        """
        new_op = self.clone()
        new_op.load_defaults()
        for nm,il in self.input_locator.items():
            new_il = InputLocator()
            new_il.tp = copy.copy(il.tp)
            new_il.val = copy.copy(il.val)
            if il.tp == entire_workflow:
                if self.inputs[nm]:
                    new_wf = self.inputs[nm].clone_wf()
                    new_op.inputs[nm] = new_wf 
            else: 
                # NOTE: have to implement __deepcopy__
                # for whatever the input is.
                new_op.inputs[nm] = copy.deepcopy(self.inputs[nm]) 
            new_op.input_locator[nm] = new_il
        #new_op.message_callback = self.message_callback
        #new_op.data_callback = self.data_callback
        return new_op

    def setup_dict(self):
        op_modulename = self.__module__[self.__module__.find('operations'):]
        op_modulename = op_modulename[op_modulename.find('.')+1:]
        dct = OrderedDict() 
        dct['op_module'] = op_modulename
        inp_dct = OrderedDict() 
        for nm,il in self.input_locator.items():
            inp_dct[nm] = {'tp':copy.copy(il.tp),'val':copy.copy(il.val)}
        dct['inputs'] = inp_dct 
        return dct
    
    def clear_outputs(self):
        for k,v in self.outputs.items():
            self.outputs[k] = None

    def description(self):
        """
        self.description() returns a string 
        documenting the input and output structure 
        and usage instructions for the Operation
        """
        return str(type(self).__name__+": "
        + self.doc_as_string()
        + "\n\n--- Inputs ---"
        + self.input_description() 
        + "\n\n--- Outputs ---"
        + self.output_description())

    def doc_as_string(self):
        if self.__doc__:
            return re.sub("\s\s+"," ",self.__doc__.replace('\n','')) 
        else:
            return "none"

    def input_description(self):
        a = ""
        for name in self.inputs.keys(): 
            if self.input_locator[name]:
                display_val = self.input_locator[name]
            else:
                display_val = self.inputs[name] 
            a = a + '\n\n'+parameter_doc(name,display_val,self.input_doc[name])
        return a

    def output_description(self):
        a = ""
        for name,val in self.outputs.items(): 
            a = a + '\n\n'+parameter_doc(name,val,self.output_doc[name])
        return a
    

def parameter_doc(name,value,doc):
    if isinstance(value, InputLocator):
        tp_str = input_types[value.tp]
        v_str = str(value.val)
        return "- name: {} \n- value: {} ({}) \n- doc: {}".format(name,v_str,tp_str,doc) 
    else:
        v_str = str(value)
        return "- name: {} \n- value: {} \n- doc: {}".format(name,v_str,doc) 
                



# -- coding: utf-8 --
"""
Integrate an ROI on image in X-direction
"""

import numpy as np
# TODO: this shouldn't depend on xicam's pipeline module
#from pipeline import remesh_integ

from ... import Operation as opmod 
from ...Operation import Operation

class RemeshXIntegration(Operation):
    """
    Input image data (ndarray), mask, ROI mask, qvrt, qpar
    Output arrays containing q and I(q) 
    """
    def __init__(self):
        input_names = list(['data', 'mask', 'ROI_mask', 'qvrt', 'qpar'])
        output_names = list(['qx','xprofile'])
        super(RemeshXIntegration, self).__init__(input_names, output_names)
        self.input_doc['data'] = '2d array representing intensity for each pixel'
        self.input_doc['mask'] = '2d array for image mask, same shape as image_data'
        self.input_doc['ROI_mask'] = '2d array for ROI mask, same shape as image_data'
        self.input_doc['qvrt'] = 'q_z axis'
        self.input_doc['qpar'] = 'q_p axis'

        self.output_doc['qx'] = 'Scattering vector magnitude q in 1/Angstrom.'
        self.output_doc['xprofile'] = 'Integrated intensity at q.'

    def run(self):
        data = self.inputs['data']
        mask = self.inputs['mask']
        cut = self.inputs['ROI_mask']
        qv = self.inputs['qvrt']
        qp = self.inputs['qpar']
        #qx, xprof = remesh_integ.remeshzintegrate(data, mask, cut = cut, qvrt = qv, qpar = ap)
        qx, xprof = None, None
        self.outputs['qx'] = qx
        self.outputs['xprofile'] = xprof



from __future__ import print_function
import os
import pkgutil
import importlib
from collections import OrderedDict

import yaml

from .. import pawstools

# check for an ops.cfg (yaml) file
cfg_file = os.path.join(pawstools.paws_cfg_dir,'ops.cfg')
if os.path.exists(cfg_file):
    load_flags = pawstools.load_cfg(cfg_file)
else:
    load_flags = OrderedDict() 

# Keep track of keys that get loaded in this run.
# These keys are used to remove load_flags automatically
# in case Operations or categories have been renamed or removed.
load_keys = []
    
def save_config():
    """
    Call save_config() before closing 
    to save the state of which ops are enabled/disabled.
    """
    # Order the load flags using load_keys...
    od_load_flags = OrderedDict()
    for k in load_keys:
        od_load_flags[k] = load_flags[k]
    pawstools.save_cfg(od_load_flags,cfg_file)

def load_ops_from_path(path_,pkg,cat_root=''):
    ops = []
    cats = []
    # pkgutil.iter_modules returns module_loader, module_name, ispkg forall modules in path
    mods = pkgutil.iter_modules(path_)
    mods = [mod for mod in mods if mod[1] not in ['__init__','Operation','OpManager','optools','DMZ']]
    for modloader, modname, ispkg in mods:
        if cat_root == '':
            mod_root = modname 
        else:
            mod_root = cat_root+'.'+modname
        if not ispkg:
            # Only load non-packages
            load_keys.append(mod_root)
        if mod_root in load_flags.keys():
            load_mod = bool(load_flags[mod_root])
        else:
            # NOTE: This line determines whether or not 
            # newly arrived modules should be loaded by default.
            load_mod = False
        load_flags[mod_root] = load_mod
        # if it is a package, recurse
        if ispkg:
            pkg_path = [os.path.join(path_[0],modname)]
            pkg_ops, pkg_cats = load_ops_from_path(pkg_path,pkg+'.'+modname,mod_root)
            pkg_ops = [op for op in pkg_ops if not op in ops]
            pkg_cats = [cat for cat in pkg_cats if not cat in cats]
            ops = ops + pkg_ops
            cats = cats + pkg_cats
        else:
            # assume that there is an Operation in this module
            # whose class name is the same as the module name.
            if not cat_root in cats:
                cats.append(cat_root)
            ops.append( (cat_root,modname) )
    return ops, cats

def disable_ops(disable_root):
    # get all keys that contain disable_root
    disable_keys = [k for k in load_flags.keys() if disable_root in k]
    for k in disable_keys:
        load_flags[k] = False 

cat_op_list, cat_list = load_ops_from_path(__path__,__name__)

load_flags = {k:load_flags[k] for k in load_flags.keys() if k in load_keys}




import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class ArrayMirrorHorizontal(Operation):
    """
    Mirror an array across a horizontal plane,
    i.e., exchange indices along axis 0.
    """

    def __init__(self):
        input_names = ['array_in']
        output_names = ['array_out']
        super(ArrayMirrorHorizontal, self).__init__(input_names, output_names)
        self.input_doc['array_in'] = '2d array'
        self.output_doc['array_out'] = 'input array mirrored horizontally'

    def run(self):
        x = self.inputs['array_in']
        if x is None:
            return
        self.outputs['array_out'] = x[::-1,:]





from __future__ import print_function
import traceback
from PySide import QtCore

from ..core.operations.Operation import Operation

class QWfWorker(QtCore.QObject):
    """
    Container for storing and executing parts of a workflow,
    to be pushed onto QtCore.QThread(s) as needed.
    """
    opDone = QtCore.Signal(str,Operation)
    allDone = QtCore.Signal()

    def __init__(self,op_dict=None,parent_QObject=None):
        super(QWfWorker,self).__init__(parent_QObject)
        self.op_dict = op_dict 

    def work(self):
        try:
            for op_tag,op in self.op_dict.items():
                op.run()
                self.opDone.emit(op_tag,op)
        except Exception as ex:
            tb = traceback.format_exc()
            msg = str('Error encountered during execution. \n'
                + 'Error message: {} \n'.format(ex.message) 
                + 'Stack trace: {}'.format(tb)) 
            # TODO: get this msg back to the main ui via signals/slots
            print(msg)
            #raise ex
        self.allDone.emit()





import os
from datetime import datetime as dt

import yaml

p = os.path.abspath(__file__)
# p = (pawsroot)/paws/core/pawstools.py

d = os.path.dirname(p)
# d = (pawsroot)/paws/core/

d = os.path.dirname(d)
# d = (pawsroot)/paws/
sourcedir = str(d)

d = os.path.dirname(d)
# d = (pawsroot)/
rootdir = str(d)

user_homedir = os.path.expanduser("~")

paws_scratch_dir = os.path.join(user_homedir,'.paws_scratch')
paws_cfg_dir = os.path.join(user_homedir,'.paws_cfg')
if not os.path.exists(paws_cfg_dir):
    os.mkdir(paws_cfg_dir)
if not os.path.exists(paws_scratch_dir):
    os.mkdir(paws_scratch_dir)

# Get the code version from the paws_config.py file, store as __version__
with open(os.path.join(sourcedir,'paws_config.py')) as f: 
    exec(f.read())
version=__version__

class WorkflowAborted(Exception):
    pass
    #def __init__(self,msg):
    #    super(WorkflowAborted,self).__init__(self,msg)

class OperationDisabledError(Exception):
    pass
    #def __init__(self,msg):
    #    super(WorkflowAborted,self).__init__(self,msg)

class WfNameError(Exception):
    pass

class PluginNameError(Exception):
    pass

class PluginLoadError(Exception):
    pass

def dtstr():
    """Return date and time as a string"""
    return dt.strftime(dt.now(),'%Y %m %d, %H:%M:%S')

def timestr():
    """Return time as a string"""
    return dt.strftime(dt.now(),'%H:%M:%S')

def save_file(filename,d):
    """
    Create or replace file indicated by filename,
    as a yaml serialization of dict d.
    """
    f = open(filename, 'w')
    yaml.dump(d, f)
    f.close()
    
def update_file(filename,d):
    """
    Save the items in dict d into filename,
    without removing members not included in d.
    """
    if os.path.exists(filename):
        f_old = open(filename,'r')
        d_old = yaml.load(f_old)
        f_old.close()
        d_old.update(d)
        d = d_old
    f = open(filename, 'w')
    yaml.dump(d, f)
    f.close()

def save_cfg(cfg_data,cfg_file):
    cfg = open(cfg_file,'w')
    yaml.dump(cfg_data,cfg)
    cfg.close()

def load_cfg(cfg_file):
    cfg = open(cfg_file,'r')
    cfg_data = yaml.load(cfg)
    cfg.close()
    if not cfg_data:
        cfg_data = OrderedDict() 
    return cfg_data




from PySide import QtGui
import qdarkstyle

def ui_app(app_args=[]):
    """
    Return a reference to a new QApplication or a currently running QApplication.
    
    Input arguments are passed to the QApplication constructor.
    If any exception is thrown,
    try to find a running QCoreApplication.
    """
    try:
        app = QtGui.QApplication(app_args)
        app.setStyleSheet(qdarkstyle.load_stylesheet() + app.styleSheet())
    except:
        app = QtGui.QApplication.instance()
    return app







import numpy as np
import pypif.obj as pifobj

from ... import Operation as opmod 
from ...Operation import Operation
from ....tools.saxs import saxs_fit 

class Pif1dScatteringIntensity(Operation):
    """Build a simple pypif.obj.ChemicalSystem containing the scattering intensity."""

    def __init__(self):
        input_names = ['uid','q_I']
        output_names = ['pif']
        super(Pif1dScatteringIntensity,self).__init__(input_names,output_names)
        self.input_doc['uid'] = 'PIF uid (string)'
        self.input_doc['q_I'] = 'n-by-2 array of q values and corresponding scattering intensities'
        self.output_doc['pif'] = 'pif object representing the input data'
        self.input_type['q_I'] = opmod.workflow_item

    def run(self):
        uid = self.inputs['uid']
        q_I = self.inputs['q_I']

        csys = pifobj.ChemicalSystem()
        csys.uid = uid
        csys.properties = []
        csys.tags = ['unlabeled scattering intensity']

        # Pack q_I into a property
        pI = self.q_I_property(q_I)
        pI.name = 'SAXS intensity' 
        csys.properties.append(pI)
        self.outputs['pif'] = csys

    def q_I_property(self,q_I):
        pI = pifobj.Property()
        n_qpoints = q_I.shape[0]
        pI.scalars = [pifobj.Scalar(q_I[i,1]) for i in range(n_qpoints)]
        pI.units = 'counts'
        pI.conditions = []
        pI.conditions.append( pifobj.Value('scattering vector', 
                            [pifobj.Scalar(q_I[i,0]) for i in range(n_qpoints)],
                            None,None,None,'1/Angstrom') )
        return pI 
        



"""Widgets for displaying text"""

from ...core.operations import Operation as opmod
from ...core.operations.Operation import Operation
from ...core.operations.optools import FileSystemIterator

unit_indent='&nbsp;&nbsp;&nbsp;&nbsp;'

def display_text(itm,indent=unit_indent):
    if type(itm).__name__ in ['str','unicode']:
        t = '(str) <br>' + indent + '{}'.format(itm)
    elif isinstance(itm,dict):
        t = '(dict)'
        for k,v in itm.items():
            t += '<br>' + indent + '{}: {}'.format(k,display_text(v,indent+unit_indent))
    elif isinstance(itm,list):
        t = '(list)'
        for i in range(len(itm)):
            t += '<br>' + indent + '{}: {}'.format(i,display_text(itm[i],indent+unit_indent))
    elif isinstance(itm,opmod.InputLocator):
        t = '(InputLocator)'
        t += '<br>' + indent + 'type: {}'.format(opmod.input_types[itm.tp])
        t += '<br>' + indent + 'val: {}'.format(itm.val)
        t += '<br>' + indent + 'data: {}'.format(type(itm.data).__name__)
    elif isinstance(itm,FileSystemIterator):
        t = '(FileSystemIterator)'
        for ip,p in zip(range(len(itm.paths_done)),itm.paths_done):
            t += '<br>' + indent + '{}: {}'.format(ip,p)
    else:
        t = '('+type(itm).__name__+')' + '<br>' + indent + '{}'.format(itm)
    return t
    
def display_text_fast(itm,indent=unit_indent):
    # TODO: Make ndarrays display truncated but without importing numpy?
    row_limit = 10
    if type(itm).__name__ in ['str','unicode']:
        t = '(str) <br>' + indent + '{}'.format(itm)
    elif isinstance(itm,dict):
        t = '(dict)'
        if len(itm) > row_limit:
            suffix = '<br>'+indent+'ETC ...'
            ndisp = row_limit 
        else:
            suffix = ''
            ndisp = len(itm)
        for k,v in itm.items()[:ndisp]:
            t += '<br>' + indent + '{}: {}'.format(k,display_text_fast(v,indent+unit_indent))
        t += suffix 
    elif isinstance(itm,list):
        t = '(list)'
        if len(itm) > row_limit:
            suffix = '<br>'+indent+'ETC ...'
            ndisp = row_limit 
        else:
            suffix = ''
            ndisp = len(itm)
        for i in range(len(itm))[:ndisp]:
            t += '<br>' + indent + '{}: {}'.format(i,display_text_fast(itm[i],indent+unit_indent))
        t += suffix 
    elif isinstance(itm,opmod.InputLocator):
        t = '(InputLocator)'
        t += '<br>' + indent + 'type: {}'.format(opmod.input_types[itm.tp])
        t += '<br>' + indent + 'val: {}'.format(itm.val)
        t += '<br>' + indent + 'data: {}'.format(type(itm.data).__name__)
    elif isinstance(itm,FileSystemIterator):
        t = '(FileSystemIterator)'
        for ip,p in zip(range(len(itm.paths_done)),itm.paths_done):
            t += '<br>' + indent + '{}: {}'.format(ip,p)
    else:
        t = '('+type(itm).__name__+')' + '<br>' + indent + '{}'.format(itm)
    return t
    




"""
Produce a PyFAI.AzumthalIntegrator to use for calibrating and integrating images.
"""

import numpy as np
import pyFAI

from ... import Operation as opmod 
from ...Operation import Operation

class BuildPyFAIIntegrator(Operation):
    """
    Input dict of calibration parameters 
    Return AzimuthalIntegrator 
    """
    def __init__(self):
        input_names = ['poni_dict']
        output_names = ['integrator']
        super(BuildPyFAIIntegrator,self).__init__(input_names,output_names)
        self.input_doc['poni_dict'] = str( 'dict of calibration parameters; '
        + 'minimally including keys dist, poni1, poni2, rot1, rot2, rot3, pixel1, pixel2, wavelength;'
        + 'optionally including keys fpolz, detector, splineFile; '
        + 'same specifications as pyFAI .poni format calibration parameters')
        self.input_type['poni_dict'] = opmod.workflow_item
        self.output_doc['integrator'] = 'PyFAI.AzimuthalIntegrator object set up with input poni_dict'

    def run(self):
        pd = self.inputs['poni_dict']
        p = pyFAI.AzimuthalIntegrator()
        if pd is not None:
            p.setPyFAI(**pd)
        self.outputs['integrator'] = p




import numpy as np
from matplotlib.figure import Figure
from scipy.special import wofz
from scipy.optimize import minimize as scimin

from ... import Operation as opmod 
from ...Operation import Operation

class VoigtPeakFit(Operation):
    """
    Fit a set of x and y values to a Voigt distribution.
    Solves the best-fitting hwhm (half width at half max) 
    of the gaussian and lorentzian distributions and shared distribution center.
    Takes as input a guess for the distribution center and hwhm.
    Range of fit is determined by weighting the objective 
    by a Hann window centered at the distribution center,
    with a window width of the distribution's estimated full width at half max.
    """

    def __init__(self):
        input_names = ['x','y','xguess','hwhm']
        output_names = ['y_voigt','xpk','ypk','hwhm_g','hwhm_l','prefactor','plot']
        super(VoigtPeakFit,self).__init__(input_names, output_names)
        self.input_doc['x'] = '1d array of x values (domain)'
        self.input_doc['y'] = '1d array of y values (amplitudes)'
        self.input_doc['xguess'] = 'initial guess for the center of the voigt profile'
        self.input_doc['hwhm'] = 'initial guess for the half width at half max of the voigt profile'
        self.input_type['x'] = opmod.workflow_item
        self.input_type['y'] = opmod.workflow_item
        self.output_doc['y_voigt'] = 'the optimized voigt profile'
        self.output_doc['xguess'] = 'optimized voigt profile distribution center'
        self.output_doc['ypk'] = 'optimized voigt profile peak amplitude'
        self.output_doc['hwhm_g'] = 'optimized voigt profile gaussian component hwhm'
        self.output_doc['hwhm_l'] = 'optimized voigt profile lorentzian component hwhm'
        self.output_doc['prefactor'] = 'optimized voigt profile multiplicative prefactor'
        self.output_doc['plot'] = 'matplotlib Figure for inspecting the fit results'

    def run(self):
        x = self.inputs['x']
        y = self.inputs['y']
        xpk = self.inputs['xguess']
        hwhm = self.inputs['hwhm']
        if x is None or y is None or xpk is None or hwhm is None:
            return
        # get y value nearest xpk guess, use it to guess a scaling factor
        ypk = y[np.argmin((x-xpk)**2)]
        scl = ypk / self.voigt(0, hwhm, hwhm)
        xpk, hwhm_g, hwhm_l, scl = self.solve_voigt(x,y,xpk,hwhm,hwhm,scl)
        mpl_fig = Figure(figsize=(100,100))
        ax = mpl_fig.add_subplot(111)
        ax.plot(x,y)
        y_voigt = self.voigt(x,hwhm_g,hwhm_l)
        ax.plot(x,y_voigt,'r')
        self.outputs['y_voigt'] = y_voigt
        self.outputs['xpk'] = xpk
        self.outputs['hwhm_g'] = hwhm_g
        self.outputs['hwhm_l'] = hwhm_l
        self.outputs['prefactor'] = scl
        self.outputs['plot'] = mpl_fig

    @staticmethod
    def solve_voigt(x, y, xc, hwhm_g, hwhm_l, scl):
        """iteratively minimize an objective to fit x, y curve to a voigt profile"""
        res = scimin(partial(self.hann_voigt_fit,x,y),(xc,hwhm_g,hwhm_l,scl))

    @staticmethod
    def hann_voigt_fit(x, y, xc, hwhm_g, hwhm_l, scl):
        # estimate hwhm of voigt
        # hwhm estimation params from https://en.wikipedia.org/wiki/Voigt_profile
        phi = hwhm_l / hwhm_g
        c0 = 2.0056; c1 = 1.0593 
        hwhm_voigt = hwhm_g * (1 - c0*c1 + np.sqrt(phi**2 + 2*c1*phi +c0**2*c1**2))
        # x,y values in the window region
        i_win = np.array([i for i in range(len(y)) if x[i] > xc-hwhm_voigt and x[i] < xc+hwhm_voigt])
        y_win = np.array([y[i] for i in i_win])
        x_win = np.array([x[i] for i in i_win])
        n_win = len(i_win)
        # window weights
        w_win = 0.5 * (1 - np.cos(2*np.pi*np.arange(n_win)/(n_win-1)) )
        # voigt profile in window, scaled by scl
        y_voigt = scl*self.voigt(x_win-xc,hwhm_g,hwhm_l)
        return np.sum(w_win * (y_voigt - y_win)**2)

    @staticmethod
    def gaussian(x, hwhm_g):
        """
        gaussian distribution at points x, 
        center 0, half width at half max hwhm_g
        """
        return np.sqrt(np.log(2)/np.pi) / hwhm_g * np.exp(-(x/hwhm_g)**2 * np.log(2))

    @staticmethod
    def lorentzian(x, hwhm_l):
        """
        lorentzian distribution at points x, 
        center 0, half width at half max hwhm_l
        """
        return hwhm_l / np.pi / (x**2+hwhm_l**2)

    @staticmethod
    def voigt(x, hwhm_g, hwhm_l):
        """
        voigt distribution resulting from convolution 
        of a gaussian with hwhm hwhm_g 
        and a lorentzian with hwhm hwhm_l
        """
        sigma = hwhm_g / np.sqrt(2 * np.log(2))
        return np.real(wofz((x+1j*hwhm_l)/sigma/np.sqrt(2))) / sigma / np.sqrt(2*np.pi)



from collections import OrderedDict

import numpy as np
from scipy import interp
from scipy.optimize import minimize

from ... import Operation as opmod 
from ...Operation import Operation
from ....tools.saxs import saxs_fit 

class SpectrumParameterization(Operation):
    """Determine approximate parameterization for a SAXS spectrum.

    The algorithm for guessing parameters
    for the size distributions of spherical nanoparticles 
    was developed and originally contributed by Amanda Fournier.    

    The inputs are a SAXS spectrum (I(q) vs. q)
    and population flags that indicate 
    what scatterer populations to parameterize.

    Any preprocessing (background subtraction, smoothing, 
    and any other corrections or cleaning)
    should be performed beforehand. 
    """

    def __init__(self):
        input_names = ['q_I', 'flags', 'fixed_params', 'fixed_param_values']
        output_names = ['params','q_I_guess']
        super(SpectrumParameterization, self).__init__(input_names, output_names)
        self.input_doc['q_I'] = 'n-by-2 array '\
            'of wave vectors(1/Angstrom) and intensities (counts)'
        self.input_doc['flags'] = 'dict of population flags, similar to the output '\
            'of paws.core.tools.saxs.saxs_classify.SaxsClassifier.classify()'
        self.input_doc['fixed_params'] = 'list of strings indicating '\
            'which parameters should be fixed to specified values- '\
            'these correspond one-to-one with fixed_param_values.'
        self.input_doc['fixed_param_values'] = 'list of values '\
            'to which the corresponding fixed_params will be set.'
        self.output_doc['params'] = 'dict of scattering equation parameters, '\
            'similar to the input of paws.core.tools.saxs.saxs_fit.compute_saxs()'

        self.output_doc['q_I_guess'] = 'n-by-2 array of q and the intensity spectrum '\
            'corresponding to the returned scattering equation parameters'

        self.input_type['q_I'] = opmod.workflow_item
        self.input_type['flags'] = opmod.workflow_item
        self.inputs['fixed_params'] = []
        self.inputs['fixed_param_values'] = []

    def run(self):
        q_I = self.inputs['q_I']
        f = self.inputs['flags'] 
        #fix_keys = self.inputs['fixed_params']
        #fix_vals = self.inputs['fixed_param_values']
        #p_fix = {}
        #for k,v in zip(fix_keys,fix_vals):
        #    p_fix[k] = v

        if not f['bad_data']:
            p = saxs_fit.parameterize_spectrum(q_I,f)
            I_guess = saxs_fit.compute_saxs(q_I[:,0],f,p)
            q_I_guess = np.array([q_I[:,0],I_guess]).T
            self.outputs['params'] = p
            self.outputs['q_I_guess'] = q_I_guess 







from PySide import QtCore, QtGui

from ...core.operations import Operation as opmod

class WorkflowGraphView(QtGui.QScrollArea):
    
    def __init__(self,wf,parent=None):
        super(WorkflowGraphView,self).__init__(parent)
        self.setWidget(WorkflowGraphWidget(wf,self))        
        #self.setFocusPolicy(QtCore.Qt.NoFocus)
        self.setWidgetResizable(True)
        #scroll_area.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        #scroll_area.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)

    def keyPressEvent(self,evnt):
        if evnt.key() == QtCore.Qt.Key_Plus:
            self.widget().zoom_in()
        elif evnt.key() == QtCore.Qt.Key_Minus:
            self.widget().zoom_out()
        else:
            super(WorkflowGraphView,self).keyPressEvent(evnt)

class WorkflowGraphWidget(QtGui.QWidget):

    def __init__(self,wf,parent=None):
        super(WorkflowGraphWidget,self).__init__(parent)        
        self.wf = wf
        # rendering scale:
        self._scale = float(1)
        # geometrical factors responsive to self._scale:
        # margins between adjacent objects
        self.hspace = 80
        self.vspace = 40
        # letter width factor?
        self.lwidth = 12 
        # letter height factor?
        self.lheight = 24 
        # letter point size?
        self.lptsize = 12 
        #self.wf.dataChanged.connect(self.update_region_by_idx) 
        self.op_coords = {}
        self.inp_coords = {}
        self.out_coords = {}
        self.update_coords()
        #h_policy = QtGui.QSizePolicy.Minimum
        #v_policy = QtGui.QSizePolicy.MinimumExpanding 
        #v_policy = QtGui.QSizePolicy.Minimum 
        #self.setSizePolicy(h_policy,v_policy)

    #@QtCore.Slot(float)
    def set_scale(self,scl):
        self._scale = float(scl)
        self.update_coords()

    #@QtCore.Slot()
    def zoom_in(self):
        self.set_scale(self._scale*1.2)

    #@QtCore.Slot()
    def zoom_out(self):
        self.set_scale(self._scale/1.2)

    def update_coords(self):
        stk,diag = self.wf.execution_stack()
        self.op_coords, self.inp_coords, self.out_coords = self.get_op_coords(stk)
        if any(self.op_coords):
            stk_height = max([coords[1][1] for coords in self.op_coords.values()]) + self._scale*self.vspace
            stk_width = max([coords[1][0] for coords in self.op_coords.values()]) + self._scale*self.hspace
            self.wf_width = stk_width
            self.wf_height = stk_height 
            self.setMinimumSize(QtCore.QSize(self.wf_width,self.wf_height))
            self.repaint()

    def get_op_coords(self,stk):
        c = {}
        inp_c = {}
        out_c = {}
        # these h,v coords will track the top left corner 
        # of each operation while iterating through the stack.
        h = self._scale*self.hspace 
        v = self._scale*self.vspace 
        for lst in stk:
            v = self._scale*self.vspace
            layer_width = 0
            for name in lst:
                op = self.wf.get_data_from_uri(name)
                d = self.op_dims(op)
                c[name] = [(h,v),(h+d[0],v+d[1])]
                v += d[1] + self._scale*self.vspace
                layer_width = max([layer_width,d[0]])
            h += layer_width + self._scale*self.hspace
        return c, inp_c, out_c

    def op_dims(self,op):
        # horizontal extent determined by longest in/out names
        hdim = max([ self.lwidth*max([len(name) for name in op.inputs.keys()] + [len(name) for name in op.outputs.keys()])+10,100])
        # vertical extent determined by number of ins and outs
        vdim = (len(op.inputs)+len(op.outputs))*self.lheight
        return (self._scale*hdim,self._scale*vdim)

    def paintEvent(self,evnt):

        #self.update_coords()

        # create a painter 
        p = QtGui.QPainter(self)
        p.setRenderHint(QtGui.QPainter.Antialiasing)
        # give it a white pen 
        pen_w = QtGui.QPen()
        q_white = QtGui.QColor(255,255,255,255)
        pen_w.setColor(q_white)
        pen_w.setWidthF(self._scale)
        p.setPen(pen_w)
        f = QtGui.QFont()
        #print self.lptsize*self._scale
        #f.setPixelSize(self.lpxsize*self._scale)
        f.setPointSize(self.lptsize*self._scale)
        p.setFont(f)
 
        # TODO: Use this to make paintEvent more efficient.
        # QPaintEvent.region() specifies the QtGui.QRegion to paint
        #paint_region = evnt.region()

        for name,coords in self.op_coords.items():
            #print 'coords of {} are topleft = {}, bottomright = {}'.format(name,coords[0],coords[1])
            topleft = QtCore.QPoint(int(coords[0][0]),int(coords[0][1]))
            bottomright = QtCore.QPoint(int(coords[1][0]),int(coords[1][1]))
            op_rec = QtCore.QRectF(topleft,bottomright)
            p.drawRect(op_rec)
            title_rec = QtCore.QRectF(  QtCore.QPoint(int(coords[0][0]),int(coords[0][1])-self.lheight*self._scale),
                                        QtCore.QPoint(int(coords[1][0]),int(coords[0][1])) )
            p.drawText(title_rec,QtCore.Qt.AlignLeft,name)
        # Headers for input and output sides
        #inphdr = QtCore.QRectF(QtCore.QPoint(-1*(recthorz+30),-1*(rectvert+10)),
        #                        QtCore.QPoint(-1*(recthorz+10),-1*rectvert))
        #outhdr = QtCore.QRectF(QtCore.QPoint(recthorz+10,-1*(rectvert+10)),
        #                        QtCore.QPoint(recthorz+30,-1*rectvert))
        #outhdr = QtCore.QRectF(QtCore.QPoint(70,-90),QtCore.QPoint(90,-80))
        #f.setUnderline(True)
        #p.setFont(f)
        #p.drawText(inphdr,QtCore.Qt.AlignCenter,op.inputs_tag)
        #p.drawText(outhdr,QtCore.Qt.AlignCenter,op.outputs_tag)
        #f.setUnderline(False)
        #p.setFont(f)
        # Label the inputs
        #n_inp = len(self.op.inputs)
        #ispc = 2*rectvert/(2*n_inp) 
        #vcrd = -1*rectvert+ispc
        #for name in self.op.inputs.keys():
        #    il = self.op.input_locator[name]
        #    rec = QtCore.QRectF(QtCore.QPoint(-1*(recthorz-10),vcrd-5),QtCore.QPoint(0,vcrd+5))
        #    p.drawText(rec,QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter,name)
        #    p.drawLine(QtCore.QPoint(-1*(recthorz-5),vcrd),QtCore.QPoint(-1*(recthorz+10),vcrd))
        #    p.drawLine(QtCore.QPoint(-1*(recthorz+10),vcrd-10),QtCore.QPoint(-1*(recthorz+10),vcrd+10))
        #    ilrec = QtCore.QRectF(QtCore.QPoint(-100,vcrd-10),QtCore.QPoint(-1*(recthorz+12),vcrd+10))
        #    p.drawText(ilrec,QtCore.Qt.AlignRight|QtCore.Qt.AlignVCenter,#|QtCore.Qt.TextWordWrap,
        #    'type: {} \nvalue: {}'.format(op.input_types[il.tp],il.val))
        #    vcrd += 2*ispc
        ## Label the outputs
        #n_out = len(self.op.outputs)
        #ispc = 2*rectvert/(2*n_out)
        #vcrd = -1*rectvert+ispc
        #for name,val in self.op.outputs.items():
        #    rec = QtCore.QRectF(QtCore.QPoint(0,vcrd-5),QtCore.QPoint(recthorz-10,vcrd+5))
        #    p.drawText(rec,QtCore.Qt.AlignRight|QtCore.Qt.AlignVCenter,name)
        #    p.drawLine(QtCore.QPoint(recthorz-5,vcrd),QtCore.QPoint(recthorz+10,vcrd))
        #    p.drawLine(QtCore.QPoint(recthorz+10,vcrd-10),QtCore.QPoint(recthorz+10,vcrd+10))
        #    outrec = QtCore.QRectF(QtCore.QPoint(recthorz+12,vcrd-10),QtCore.QPoint(100,vcrd+10))
        #    p.drawText(outrec,QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter,type(val).__name__)
        #    #p.drawText(outrec,QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter,str(val))
        #    #|QtCore.Qt.TextWordWrap,str(val))
        #    vcrd += 2*ispc
       
    #def repaint_region_by_idx(self,topleft_idx,bottomright_idx):
    #    # assume topleft and bottomright are the same,
    #    # i.e. assume the workflow calls dataChanged()
    #    # on just one TreeItem at a time
    #    if not topleft_idx == bottomright_idx:
    #        msg = str('[{}] repaint event was called for a region. '.format(__name__)
    #        + 'Currently only single-index regions can be repainted.')
    #        raise ValueError(msg)
    #    idx = topleft_idx
    #    itm = self.get_item(idx)
    #    # get the coords for this item, call self.update() on that region
    

    #def update(self):




import os

import numpy as np
import pyFAI

from .. import Operation as opmod 
from ..Operation import Operation

class ReadPONI(Operation):
    """
    Reads in a .poni file as output by 
    pyFAI.geometry.Geometry.save(),
    outputs a poni dict
    as produced by pyFAI.geometry.Geometry.getPyFAI().
    """
    
    def __init__(self):
        input_names = ['poni_file']
        output_names = ['poni_dict']
        super(ReadPONI,self).__init__(input_names,output_names)
        self.input_doc['poni_file'] = 'path to the .poni file' 
        self.output_doc['poni_dict'] = 'Dict of pyFAI calibration parameters'

    def run(self):
        fpath = self.inputs['poni_file']
        if fpath is None:
            return
        g = pyFAI.geometry.Geometry()
        g.read(fpath)
        self.outputs['poni_dict'] = g.getPyFAI() 




from __future__ import print_function
from collections import OrderedDict
from functools import partial
import copy
import os

from PySide import QtCore

from .QWfWorker import QWfWorker
from ..core.workflow.Workflow import Workflow
from ..core.operations import Operation as opmod
from .QTreeSelectionModel import QTreeSelectionModel
from ..core.operations import optools
from . import qttools
from ..core.operations.Operation import Operation

class QWorkflow(Workflow,QTreeSelectionModel):
    """
    A QTreeSelectionModel representing a Workflow
    """
    wfFinished = QtCore.Signal()
    emitMessage = QtCore.Signal(str)
    opFinished = QtCore.Signal(str)
    emitData = QtCore.Signal(str,object)

    def __init__(self):
        super(QWorkflow,self).__init__()

    @QtCore.Slot(str,str,object)
    def relayOpData(self,op_tag,data_uri,data):
        self.emitData.emit(op_tag+'.'+data_uri,data)

    @QtCore.Slot(str)
    def relayMessage(self,msg):
        self.emitMessage.emit(msg)

    @QtCore.Slot(str,str,object)
    def updateOpInput(self,opnm,inpnm,inpdata):
        self.set_op_item(opnm,'inputs.'+inpnm,inpdata)

    @QtCore.Slot(str,str,object)
    def updateOpOutput(self,opnm,outnm,outdata):
        self.set_op_item(opnm,'outputs.'+outnm,outdata)

    @QtCore.Slot(str,str,object)
    def updateOpItem(self,opnm,item_uri,item_data):
        self.set_op_item(opnm,item_uri,item_data)

    @QtCore.Slot(str,str,object)
    def updateItem(self,item_uri,item_data):
        self.set_item(item_uri,item_data)

    def add_op(self,op_tag,op):
        op.message_callback = self.message_callback
        op.data_callback = partial( self.relayOpData,op_tag )
        self.set_item(op_tag,op)

    def headerData(self,section,orientation,data_role):
        if (data_role == QtCore.Qt.DisplayRole and section == 0):
            return "{} operation(s) loaded".format(self.root_item().n_children())
        else:
            return super(QWorkflow,self).headerData(section,orientation,data_role)    

    def execute(self):
        stk,diag = self.execution_stack()
        self.message_callback(os.linesep+'running workflow:'+os.linesep+self.print_stack(stk))
        for lst in stk:
            self.message_callback('running: {}'.format(lst))
            for op_tag in lst: 
                op = self.get_data_from_uri(op_tag) 
                for inpnm,il in op.input_locator.items():
                    if il.tp == opmod.workflow_item:
                        #op.inputs[inpnm] = self.locate_input(il)
                        self.set_op_item(op_tag,'inputs.'+inpnm,self.locate_input(il))
                        if self.data_callback:
                            self.data_callback(op_tag+'.inputs.'+inpnm,op.inputs[inpnm])
                op.run()
                for outnm,outdata in op.outputs.items():
                    if self.data_callback:
                        out_uri = op_tag+'.outputs.'+outnm
                        if outdata is not None:
                            self.data_callback(out_uri,outdata)
                        #print('done setting {}'.format(out_uri))
                self.opFinished.emit(op_tag)
                    #self.record_op_output(op_tag,outnm,outdata)
                #self.set_item(op_tag,op)
                #self.opChanged.emit(op_tag,op)
        self.wfFinished.emit()




"""This package defines widgets that are used to communicate with paws."""

import numpy as np
from PySide import QtGui
from matplotlib.figure import Figure
#from pypif.obj import System

from ...core.operations.Operation import Operation
from ...core.workflow.Workflow import Workflow 
from ...core.plugins.PawsPlugin import PawsPlugin
from .. import qttools
from .OpWidget import OpWidget
#from .PifWidget import PifWidget
from .WorkflowGraphView import WorkflowGraphView
from .text_widgets import display_text, display_text_fast

if qttools.have_qt47:
    from . import plotmaker_pqg as plotmaker
else:
    from . import plotmaker_mpl as plotmaker

def make_widget(itm):
    #if isinstance(itm,System):
    #    w = PifWidget(itm)
    if isinstance(itm,Workflow):
        w = WorkflowGraphView(itm)
    elif isinstance(itm,PawsPlugin):
        w = QtGui.QTextEdit()
        msg = str('selected plugin is a {}. '.format(type(itm).__name__)
                + 'There are no display methods associated with this plugin. '
                + 'Add a display method to {} to view this plugin.'.format(__name__))
        w.setText(msg)
    elif isinstance(itm,Operation):
        w = OpWidget(itm)
    elif isinstance(itm,np.ndarray):
        #if itm.dtype == np.int_ or itm.dtype == np.float_:
        dims = np.shape(itm)
        try:
            if len(dims) == 2 and dims[0] > 2 and dims[1] > 2:
                w = plotmaker.array_plot_2d(itm)
            elif len(dims) == 1 or (len(dims) == 2 and (dims[0]==2 or dims[1]==2)):
                w = plotmaker.array_plot_1d(itm)
        except:
            t = display_text_fast(itm)
            w = QtGui.QTextEdit(t)
        #else:
        #    t = display_text_fast(itm)
        #    w = QtGui.QTextEdit(t)
    elif isinstance(itm,Figure):
        w = plotmaker.plot_mpl_fig(itm)
    elif (isinstance(itm,float) 
    or isinstance(itm,int) 
    or isinstance(itm,dict) 
    or isinstance(itm,list) 
    or isinstance(itm,str)
    or isinstance(itm,unicode)):    
        t = display_text_fast(itm)
        w = QtGui.QTextEdit(t)
    else:
        msg = str('[{}]: selected item ({}) is not supported in {}'
            .format(__name__,type(itm).__name__,__name__)
            + '<br><br>Printout of item: <br>{}'.format(itm))
        w = QtGui.QTextEdit(msg)
    return w







import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation

class MovingAverage(Operation):
    """
    Applies moving average smoothing filter to 1d array,
    optionally weighted by window shape and error values.
    """

    def __init__(self):
        input_names = ['data','window','shape','error']
        output_names = ['smoothdata']
        super(MovingAverage, self).__init__(input_names, output_names)
        self.input_doc['data'] = '1d array'
        self.input_doc['window'] = 'integer number of data points to average on either side'
        self.input_doc['shape'] = 'window shape for weighting- triangular or square (default)'
        self.input_doc['error'] = '1d array, same shape as data, optional (default None)'
        self.output_doc['smoothdata'] = 'smoothed 1d array'
        self.input_type['data'] = opmod.workflow_item
        self.inputs['window'] = 3
        self.inputs['shape'] = 'square' 

    def run(self):
        x = self.inputs['data']
        if x is None:
            return
        w = self.inputs['window']
        err = self.inputs['error']
        nx = len(x)
        if self.inputs['shape'] == 'triangle': 
            shape_weights = (w+1-np.arange(w+1, dtype=float))/float(w+1)
        else:
            shape_weights = np.ones(w+1, dtype=float)
        shape_weights = np.concatenate( (shape_weights[::-1],shape_weights[1:]))
        if err is not None:
            err_weights = err**-2
        else:
            err_weights = np.ones(x.shape, dtype=float)
        x_out = np.array([np.sum( 
                            np.array(x[max(0,i-w):min(i+w,nx)],dtype=float) 
                             * np.array(err_weights[max(0,i-w):min(i+w,nx)]) 
                             * np.array(shape_weights[max(0,w-i):min(2*(nx-i)+1,2*w+1)]) )/
                            float(min(i+w,nx-1)-max(0,i-w)) 
                            for i in range(nx) ])
        self.outputs['smoothdata'] = x_out




"""
Various tools for working with Workflows and Operations
"""

import glob
import copy
from collections import Iterator
from collections import OrderedDict

from . import Operation as opmod 

class FileSystemIterator(Iterator):

    def __init__(self,dirpath,regex,include_existing_files=True):
        self.dirpath = dirpath
        self.rx = regex
        self.paths_done = []
        if not include_existing_files:
            self.paths_done = glob.glob(self.dirpath+'/'+self.rx)
        super(FileSystemIterator,self).__init__()

    def next(self):
        batch_list = glob.glob(self.dirpath+'/'+self.rx)
        for path in batch_list:
            if not path in self.paths_done:
                self.paths_done.append(path)
                return path
        return None

class ExecutionError(Exception):
    def __init__(self,msg):
        super(ExecutionError,self).__init__(self,msg)

def get_uri_from_dict(uri,d):
    keys = uri.split('.')
    itm = d
    for k in keys:
        if not isinstance(itm,dict):
            msg = 'something in {} is not a dict'.format(uri)
            raise KeyError(msg)
        if not k in itm.keys():
            msg = 'did not find uri {} in dict'.format(uri)
            raise KeyError(msg)
        else:
            itm = itm[k]
    return itm

def dict_contains_uri(uri,d):
    keys = uri.split('.')
    itm = d
    for k in keys:
        if not k in itm.keys():
            return False
        else:
            itm = itm[k]
    return True




"""
Configuration flags, widgets, and functions for the paws qt layer
"""

from __future__ import print_function 
import os
import re
import platform
from functools import partial

from PySide import QtGui, QtCore, QtUiTools
import yaml

from ..core.operations import Operation as opmod
from ..core import pawstools
from ..core.models.ListModel import ListModel

## Test whether we have Qt >= 4.7 
have_qt47 = True
versionReq = [4, 7]
QtVersion = QtCore.__version__ 
m = re.match(r'(\d+)\.(\d+).*', QtVersion)
if m is not None and list(map(int, m.groups())) < versionReq:
    have_qt47 = False

## Test whether we are using Windows
if platform.system() == 'Windows':
    have_windows = True
else:
    have_windows = False

class RunnableExecutor(QtCore.QRunnable):
    """
    QRunnable that handles execution of a QWorkflow
    """

    def __init__(self,wf):
        super(RunnableExecutor,self).__init__()
        self.wf = wf

    def run(self):
        self.wf.execute()

class QSourceEdit(QtGui.QTextEdit):
    
    def __init__(self):
        super(QSourceEdit,self).__init__()

    def keyPressEvent(self,evnt):
        if evnt.key() == QtCore.Qt.Key_Tab:
            # insert 4 spaces
            # TODO: Make this a user-configurable thing?
            super(QSourceEdit,self).insertPlainText('    ')
        else:
            super(QSourceEdit,self).keyPressEvent(evnt)

def text_widget(text):
    """
    Produce a Read-only Center-aligned QtGui.QLineEdit from input text.
    """
    widg = QtGui.QLineEdit(text)
    widg.setReadOnly(True)
    widg.setAlignment(QtCore.Qt.AlignHCenter)
    return widg 

def toggle_expand(trview,idx):
    trview.setExpanded(idx, not trview.isExpanded(idx))

def type_selection_widget(src,widg=None):
    if not widg:
        widg = QtGui.QComboBox()
        lm = ListModel(opmod.input_types,widg)
        widg.setModel(lm)
    else:
        lm = ListModel(opmod.input_types,widg)
        widg.setModel(lm)
    for tp in opmod.valid_types:
        lm.set_enabled(tp)
    for tp in opmod.invalid_types[src]:
        lm.set_disabled(tp)
        #widg.model().set_disabled(tp)
    return widg 

def r_hdr_widget(text):
    widg = QtGui.QLineEdit(text)
    widg.setReadOnly(True)
    widg.setAlignment(QtCore.Qt.AlignRight)
    widg.setStyleSheet( "QLineEdit { background-color: transparent }" + widg.styleSheet() )
    return widg 

def hdr_widget(text):
    widg = QtGui.QLineEdit(text)
    widg.setReadOnly(True)
    widg.setAlignment(QtCore.Qt.AlignLeft)
    widg.setStyleSheet( "QLineEdit { background-color: transparent }" + widg.styleSheet() )
    return widg 

def smalltext_widget(text):
    widg = text_widget(text)
    widg.setMaximumWidth(20*len(text))
    widg.setStyleSheet( "QLineEdit { background-color: transparent }" + widg.styleSheet() )
    return widg

def name_widget(name):
    name_widget = QtGui.QLineEdit(name)
    name_widget.setReadOnly(True)
    name_widget.setAlignment(QtCore.Qt.AlignRight)
    ht = name_widget.sizeHint().height()
    name_widget.sizeHint = lambda: QtCore.QSize(10*len(name_widget.text()),ht)
    name_widget.setSizePolicy(QtGui.QSizePolicy.Minimum,QtGui.QSizePolicy.Fixed)
    return name_widget

def bigtext_widget(text=None):
    #trunc_limit = 200
    #if len(text) > trunc_limit:
    #    text = text[:trunc_limit]+'...'
    if text is None:
        text = ''
    widg = QtGui.QLineEdit(text)
    widg.setReadOnly(True)
    widg.setAlignment(QtCore.Qt.AlignLeft)
    ht = widg.sizeHint().height()
    widg.sizeHint = lambda: QtCore.QSize(20*len(text),ht)
    #widg.setSizePolicy(QtGui.QSizePolicy.Maximum,QtGui.QSizePolicy.Fixed)
    widg.setSizePolicy(QtGui.QSizePolicy.Expanding,QtGui.QSizePolicy.Fixed)
    return widg

def start_save_ui(parent,fspath=None):
    ui_file = QtCore.QFile(os.path.join(pawstools.sourcedir,'qt','qtui','save_browser.ui'))
    ui_file.open(QtCore.QFile.ReadOnly)
    save_ui = QtUiTools.QUiLoader().load(ui_file)
    ui_file.close()
    # set ui to be deleted and to emit destroyed() signal when its window is closed
    save_ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
    trmod = QtGui.QFileSystemModel()
    trmod.setNameFilters(['*.wfl'])
    #save_ui.tree_box.setTitle('Select or enter a .wfl file to save current workflow'.format(wfname))
    save_ui.tree.setModel(trmod)
    save_ui.tree.hideColumn(1)
    save_ui.tree.hideColumn(3)
    save_ui.tree.setColumnWidth(0,400)
    save_ui.tree.expandAll()
    save_ui.tree.clicked.connect( partial(save_path,save_ui) )
    save_ui.tree.expanded.connect( partial(save_path,save_ui) )
    trmod.setRootPath(QtCore.QDir.currentPath())
    idx = trmod.index(QtCore.QDir.currentPath())
    if fspath is not None:
        try:
            trmod.setRootPath(fspath)
            idx = trmod.index(fspath)
        except:
            pass
    while idx.isValid():
        save_ui.tree.setExpanded(idx,True)
        idx = idx.parent()
    save_ui.setParent(parent,QtCore.Qt.Window)
    save_ui.setWindowModality(QtCore.Qt.ApplicationModal)
    save_ui.save_button.setText('&Save')
    #save_ui.filename.returnPressed.connect(partial(stop_save_ui,save_ui))
    save_ui.filename.textChanged.connect( partial(toggle_save_button,save_ui) )
    save_ui.filename.setText(trmod.rootPath())
    return save_ui

def toggle_save_button(ui,txt):
    idx = ui.tree.model().index(ui.filename.text())
    if idx.isValid() and ui.tree.model().isDir(idx):
        ui.save_button.setEnabled(False)
    else:
        ui.save_button.setEnabled(True)

def save_path(ui,idx=QtCore.QModelIndex(),oldidx=QtCore.QModelIndex()):
    if idx.isValid():
        p = ui.tree.model().filePath(idx)
        ui.filename.setText(p)

def start_load_ui(parent,fspath=None):
    ui_file = QtCore.QFile(os.path.join(pawstools.sourcedir,'qt','qtui','load_browser.ui'))
    ui_file.open(QtCore.QFile.ReadOnly)
    load_ui = QtUiTools.QUiLoader().load(ui_file)
    ui_file.close()
    # set ui to be deleted and to emit destroyed() signal when its window is closed
    load_ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
    trmod = QtGui.QFileSystemModel()
    trmod.setNameFilters(['*.wfl'])
    #load_ui.tree_box.setTitle('Select a .wfl file from which to load a workflow')
    load_ui.tree.setModel(trmod)
    load_ui.tree.hideColumn(1)
    load_ui.tree.hideColumn(3)
    load_ui.tree.setColumnWidth(0,400)
    load_ui.tree.expandAll()
    load_ui.tree.clicked.connect( partial(load_path,load_ui) )
    trmod.setRootPath(QtCore.QDir.currentPath())
    idx = trmod.index(QtCore.QDir.currentPath())
    if fspath is not None:
        try:
            trmod.setRootPath(fspath)
            idx = trmod.index(fspath)
        except:
            pass
    while idx.isValid():
        load_ui.tree.setExpanded(idx,True)
        idx = idx.parent()
    load_ui.setParent(parent,QtCore.Qt.Window)
    #load_ui.setAttribute(QtCore.Qt.WA_OpaquePaintEvent)
    load_ui.setWindowModality(QtCore.Qt.ApplicationModal)
    load_ui.load_button.setText('&Load')
    return load_ui

def load_path(ui,idx=QtCore.QModelIndex()):
    if idx.isValid():
        p = ui.tree.model().filePath(idx)
        if (ui.tree.model().isDir(idx) 
        or not os.path.splitext(p)[1] == '.wfl'):
            ui.load_button.setEnabled(False)
        else:
            ui.load_button.setEnabled(True)
    
def message_ui(parent):
    ui_file = QtCore.QFile(os.path.join(pawstools.sourcedir,'qt','qtui','message.ui'))
    ui_file.open(QtCore.QFile.ReadOnly)
    msg_ui = QtUiTools.QUiLoader().load(ui_file)
    ui_file.close()
    msg_ui.setParent(parent,QtCore.Qt.Window)
    msg_ui.setAttribute(QtCore.Qt.WA_OpaquePaintEvent)
    msg_ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
    msg_ui.setWindowModality(QtCore.Qt.WindowModal)
    #msg_ui.setMaximumHeight(200)
    msg_ui.message_box.setReadOnly(True)
    msg_ui.ok_button.setText('OK')
    msg_ui.ok_button.clicked.connect(msg_ui.close)
    msg_ui.ok_button.clicked.connect(msg_ui.deleteLater)
    msg_ui.ok_button.setFocus()
    msg_ui.ok_button.setDefault(True)
    return msg_ui






from PySide import QtCore

from ..core.models.TreeModel import TreeModel

class QTreeModel(TreeModel,QtCore.QAbstractItemModel):
    """
    A Qt Model-View interface for a TreeModel.
    Required virtual methods: index(), parent(), rowCount(), columnCount(), and data().
    Resizeable TreeModels should implement
    insertRows(), removeRows(), insertColumns(), and removeColumns().
    To customize the header in QAbstractItemViews, implement headerData().
    """

    def __init__(self,flag_dict):
        super(QTreeModel,self).__init__(flag_dict)

    # Subclass of QAbstractItemModel must implement index()
    def index(self,row,col,p_idx):
        """
        Returns QModelIndex address of int row, int col, under QModelIndex p_idx.
        If a row, column, p_idx combination is invalid, return QModelIndex().
        """
        if not p_idx.isValid():
            if row == 0:
                return self.root_index()
            else:
                return QtCore.QModelIndex()
        else:
            p_itm = self.get_from_index(p_idx)
            if row < p_itm.n_children() and row >= 0:
                return self.createIndex(row,col,p_itm.children[row])
            else:
                return QtCore.QModelIndex()

    def root_index(self):
        return self.createIndex(0,0,self.root_item())

    def root_item(self):
        return self._root_item

    def get_from_index(self,idx=QtCore.QModelIndex()):
        """
        For a valid QModelIndex, return idx.internalPointer().
        For an invalid index, return None.
        """
        if idx.isValid():
            return idx.internalPointer() 
        else:
            return None 

    def get_data_from_index(self,idx=QtCore.QModelIndex()):
        uri = self.get_uri_of_index(idx)
        return self.get_data_from_uri(uri)

    def get_uri_of_index(self,idx):
        itm = self.get_from_index(idx)
        return self.build_uri(itm)

    def get_index_of_item(self,itm):
        uri = itm.build_uri()
        return self.get_index_of_uri(uri)

    def get_index_of_uri(self,uri):
        path = uri.split('.')
        idx = self.root_index()
        for k in path[:-1]:
            itm = self.get_from_index(idx)
            child_keys = [c.tag for c in itm.children]
            idx = self.index(child_keys.index(k),0,idx)
        k = path[-1]
        if k:
            itm = self.get_from_index(idx)
            child_keys = [c.tag for c in itm.children]
            return self.index(child_keys.index(k),0,idx)
        elif k == '':
            return idx 

    def set_item_at_uri(self,itm_uri,itm_data):
        if '.' in itm_uri:
            parent_uri = itm_uri[:itm_uri.rfind('.')]
            parent_idx = self.get_index_of_uri(parent_uri)
            itm_tag = itm_uri[itm_uri.rfind('.')+1:]
        else:
            parent_idx = self.root_index()
            itm_tag = itm_uri
        self.set_item(itm_tag,itm_data,parent_idx)        

    #def set_item(self,itm_tag,itm_data=None,parent_idx=None):
    #    if parent_idx is None:
    #        parent_idx = self.root_index()
    #    parent_itm = self.get_from_index(parent_idx)
    #    parent_uri = self.get_uri_of_index(parent_idx)
    #    if parent_uri:
    #        itm_uri = parent_uri+'.'+itm_tag
    #    else:
    #        itm_uri = itm_tag
    #    # add TreeItems to index the new model content:
    #    treedata = self.build_tree(itm_data)
    #    self.tree_update(parent_idx,itm_tag,treedata)
    #    # store the data:
    #    self.set_item(itm_uri,itm_data)
    #    # signal dataChanged down the tree
    #    child_keys = [c.tag for c in parent_itm.children]
    #    itm_idx = self.index(child_keys.index(itm_tag),0,parent_idx)
    #    self.tree_dataChanged(itm_idx) 

    def tree_update_at_uri(self,itm_uri,itm_data):
        if '.' in itm_uri:
            parent_uri = itm_uri[:itm_uri.rfind('.')]
            parent_idx = self.get_index_of_uri(parent_uri)
            itm_tag = itm_uri[itm_uri.rfind('.')+1:]
        else:
            parent_idx = self.root_index()
            itm_tag = itm_uri
        self.tree_update(parent_itm,itm_tag,self.build_tree(itm_data))        

    def tree_update(self,parent_itm,itm_tag,treedata):
        child_keys = [c.tag for c in parent_itm.children]
        parent_idx = self.get_index_of_item(parent_itm)
        if itm_tag in child_keys:
            # Find existing itm under parent.
            itm_row = child_keys.index(itm_tag)
            itm = parent_itm.children[itm_row]
            idx = self.index(itm_row,0,parent_idx)
            # Remove any children that do not represent the new itm_tree data
            new_keys = []
            if isinstance(treedata,dict):
                new_keys = treedata.keys()
            for gc_row in range(itm.n_children())[::-1]:
                gc_itm = itm.children[gc_row]
                if not gc_itm.tag in new_keys:
                    #rm_row = [gc.tag for gc in itm.children].index(grandchild_itm.tag)
                    self.beginRemoveRows(idx,gc_row,gc_row)
                    itm.children.pop(gc_row) 
                    self.endRemoveRows()
        else:
            # Else, put a new TreeItem at the end row
            itm_row = parent_itm.n_children()
            itm = self.create_tree_item(parent_itm,itm_tag)
            self.beginInsertRows(parent_idx,itm_row,itm_row)
            parent_itm.children.insert(itm_row,itm)
            self.endInsertRows()
        # If needed, get the index of the new item and recurse 
        if isinstance(treedata,dict):
            for tag,val in treedata.items():
                self.tree_update(itm,tag,val)

    def remove_item(self,itm_uri):
        itm = self.get_from_uri(itm_uri)
        parent_itm = itm.parent 
        parent_idx = self.get_index_of_item(parent_itm) 
        if parent_idx is None:
            parent_idx = self.root_index()
        child_keys = [c.tag for c in parent_itm.children]
        rm_row = child_keys.index(itm_tag)
        rm_idx = self.index(rm_row,0,parent_idx)
        self.beginRemoveRows(parent_idx,rm_row,rm_row)
        parent_itm.children.pop(rm_row)
        self.endRemoveRows()
        self.dataChanged.emit(rm_idx,rm_idx)
        self.tree_dataChanged(parent_idx)

    def tree_dataChanged(self,idx):
        self.dataChanged.emit(idx,idx)
        itm = idx.internalPointer()
        for c_row in range(itm.n_children())[::-1]:
            for c_col in range(1+len(itm.flags)):
                c_idx = self.index(c_row,c_col,idx)
                self.tree_dataChanged(c_idx)

    # Subclass of QAbstractItemModel must implement parent()
    def parent(self,idx):
        """
        Returns QModelIndex of parent of item at QModelIndex index
        """
        itm = self.get_from_index(idx)
        if itm == self.root_item():
            return QtCore.QModelIndex()
        parent_itm = itm.parent
        if parent_itm == self.root_item():
            return self.root_index()
        parent_tag = parent_itm.tag
        grandparent_itm = parent_itm.parent
        parent_keys = [p.tag for p in grandparent_itm.children]
        parent_row = parent_keys.index(parent_tag)
        return self.createIndex(parent_row,0,grandparent_itm.children[parent_row])
        
    # Subclass of QAbstractItemModel must implement rowCount()
    def rowCount(self,parent_idx=QtCore.QModelIndex()):
        """
        Either give the number of top-level items,
        or count the children of parent
        """
        if not parent_idx.isValid():
            # there is only 1 root item 
            return 1
        else:
            # count children of parent item
            parent_itm = self.get_from_index(parent_idx)
            return parent_itm.n_children()
    
    # Subclass of QAbstractItemModel must implement columnCount()
    def columnCount(self,parent=QtCore.QModelIndex()):
        """
        TreeModels by default have one column.
        More columns can be added by reimplementing columnCount()
        and then providing for them in TreeModel.data().
        """
        return 1

    # QAbstractItemModel subclass must implement 
    # data(QModelIndex[,role=Qt.DisplayRole])
    def data(self,itm_idx,data_role):
        """
        TreeModel's implementation of data() 
        returns the tag of the TreeItem at itm_idx.
        This is only if itm_idx.column() == 0.
        Subclasses can reimplement data() 
        to provide meaningful output for other columns,
        and may consider falling back on super().data() 
        if itm_idx.column() == 0.
        """
        if (not itm_idx.isValid()):
            return None
        #if itm_idx == self.root_index():
        #    return None
        #itm = itm_idx.internalPointer()
        if ((data_role == QtCore.Qt.DisplayRole
        or data_role == QtCore.Qt.ToolTipRole 
        or data_role == QtCore.Qt.StatusTipRole
        or data_role == QtCore.Qt.WhatsThisRole)
        and itm_idx.column() == 0):
            itm = self.get_from_index(itm_idx)
            return itm.tag
        else:
            return None

    def headerData(self,section,orientation,data_role):
        # note: section indicates row or column number, depending on orientation
        if (data_role == QtCore.Qt.DisplayRole and section == 0):
            return "{} item(s)".format(self.n_children())
        else:
            return None




from __future__ import print_function
import importlib
from collections import OrderedDict

from ..operations.Operation import Operation
from ..workflow.Workflow import Workflow
from ..models.TreeModel import TreeModel
from .. import plugins as pgns
from .. import pawstools
from .PawsPlugin import PawsPlugin

class PluginManager(TreeModel):
    """
    Tree structure for managing paws plugins.
    """

    def __init__(self,**kwargs):
        flag_dict = OrderedDict()
        flag_dict['select'] = False
        super(PluginManager,self).__init__(flag_dict)
        self.logmethod = print 

    def write_log(self,msg):
        self.logmethod(msg)
        #if self.logmethod:
        #else:
        #    print(msg)

    def plugin_setup_dict(self,pgin):
        pgin_mod = pgin.__module__[pgin.__module__.find('plugins'):]
        pgin_mod = pgin_mod[pgin_mod.find('.')+1:]
        dct = OrderedDict() 
        dct['plugin_module'] = pgin_mod
        dct['inputs'] = pgin.inputs 
        return dct

    def load_from_dict(self,pgin_name,pgin_spec):
        """
        Load plugins from a dict that specifies their setup parameters.
        """
        #for tag, pgin_spec in plugin_dict.items():
        pgin_module = pgin_spec['plugin_module']
        pgin = self.load_plugin(pgin_module)
        if pgin is not None:
            if not issubclass(pgin,PawsPlugin):
                self.write_log('Did not find Plugin {} - skipping.'.format(pgin_uri))
            else:
                pgin = pgin()
                for name in pgin.inputs.keys():
                    if name in pgin_spec['inputs']:
                        pgin.inputs[name] = pgin_spec['inputs'][name]
                pgin.start()
                # if already have this uri, first generate auto_tag
                #if self.tree_contains_uri(uri):
                #    uri = self.auto_tag(uri)
                #self.add_plugin(uri,pgin)
                self.set_item(pgin_name,pgin)
        else:
            self.write_log('Did not find Plugin {} - skipping.'.format(ptype))

    #def tree_update(self,parent_idx,itm_tag,itm_data):
    #    if isinstance(itm_data,Operation) or isinstance(itm_data,PawsPlugin):
    #        super(PluginManager,self).tree_update(parent_idx,itm_tag,self.build_tree(itm_data))
    #    else:
    #        super(PluginManager,self).tree_update(parent_idx,itm_tag,itm_data)

    def add_plugin(self,pgin_name,pgin):
        """
        Add a plugin, with key specified by pgin_name.
        If pgin_name is not unique (i.e. a plugin with that name already exists),
        this method will overwrite the existing plugin with a new one.
        """
        if not self.is_tag_valid(pgin_name): 
            raise pawstools.PluginNameError(self.tag_error_message(pgin_name))
        self.set_item(pgin_name,pgin)

    def build_tree(self,x):
        """
        Reimplemented TreeModel.build_tree() 
        so that TreeItems are built from PawsPlugins
        and Workflows and Operations.
        """
        if isinstance(x,PawsPlugin):
            #d = x.content()
            d = OrderedDict()
            for k,v in x.content().items():
                d[k] = self.build_tree(v)
        elif isinstance(x,Workflow):
            #d = x.op_dict()
            d = OrderedDict()
            for k,v in x.op_dict().items():
                d[k] = self.build_tree(v)
        elif isinstance(x,Operation):
            d = OrderedDict()
            d['inputs'] = self.build_tree(x.inputs)
            d['outputs'] = self.build_tree(x.outputs)
        else:
            return super(PluginManager,self).build_tree(x) 
        return d

    def get_plugin(self,pgin_tag):
        return self.get_data_from_uri(pgin_tag)

    def load_plugin(self,pgin_module):    
        try:
            mod = importlib.import_module('.'+pgin_module,pgns.__name__)
            return mod.__dict__[pgin_module]
        except Exception as ex:
            msg = str('Exception occurred while loading plugin {}. '
            .format(pgin_module) + 'Error message: ' + ex.message)
            self.write_log(msg)
            raise pawstools.PluginLoadError(msg)   
 
    def n_plugins(self):
        return self.n_children() 

    def list_plugin_tags(self):
        return self.root_tags()
        #r = self.get_from_idx(self.root_index())
        #return [itm.tag for itm in r.children]




"""
This runs various widgets built on the paws.api.
"""

import sys

from .. import qtapi

from .. import ui_app
from ..UiManager import UiManager

def main():   
    """
    An entry point for paws full-featured interface.
    """
    # start QtGui.QApplication
    app = ui_app(sys.argv)
    # start paws objects 
    corepaw = qtapi.start(app)

    # start a ui manager
    ui_manager = UiManager(corepaw)

    ### LAUNCH ###
    ui_manager.ui.show()
    ret = app.exec_()
    # TODO: connect corepaw.save_config to app.aboutToQuit() signal,
    # in case app.exec_() does not have time to return,
    # which may happen in some situations on some platforms.
    corepaw.save_config()
    sys.exit(ret)
    
# TODO: entry points for alternative/simplified interfaces, e.g.:
# def specialwidget():
#
#    Load an app and start a paw
#
#    Set up a UI for this special case
#
#    Launch 







from PySide import QtGui, QtCore

class OpWidget(QtGui.QWidget):
    
    def __init__(self,op):
        super(OpWidget,self).__init__()
        self.op = op
        #self.render_from_op()        

    def paintEvent(self,evnt):
        w = self.width()
        h = self.height()
        widgdim = float( min([w,h]) )
        # Create a painter and draw in the elements of the Operation
        p = QtGui.QPainter(self)
        p.setRenderHint(QtGui.QPainter.Antialiasing)
        pen = QtGui.QPen()
        qwhite = QtGui.QColor(255,255,255,255)
        pen.setColor(qwhite)
        p.setPen(pen)
        #p.setBrush()...
        p.translate(w/2, h/2)
        p.scale(widgdim/200,widgdim/200)
        rectvert = 80 
        recthorz = 50
        topleft = QtCore.QPoint(int(-1*recthorz),int(-1*rectvert))
        bottomright = QtCore.QPoint(int(recthorz),int(rectvert))
        # Large rectangle representing the Operation
        mainrec = QtCore.QRectF(topleft,bottomright)
        p.drawRect(mainrec)
        f = QtGui.QFont()
        title_hdr = QtCore.QRectF(QtCore.QPoint(-100,-1*(rectvert+10)),
                                QtCore.QPoint(100,-1*rectvert))
        #title_hdr = QtCore.QRectF(QtCore.QPoint(-30,-10),QtCore.QPoint(30,10))
        #f.setPixelSize(10)
        f.setPointSize(5)
        p.setFont(f)
        p.drawText(title_hdr,QtCore.Qt.AlignCenter,type(self.op).__name__)
        f.setPointSize(4)
        p.setFont(f)
        # Headers for input and output sides
        inphdr = QtCore.QRectF(QtCore.QPoint(-1*(recthorz+30),-1*(rectvert+10)),
                                QtCore.QPoint(-1*(recthorz+10),-1*rectvert))
        outhdr = QtCore.QRectF(QtCore.QPoint(recthorz+10,-1*(rectvert+10)),
                                QtCore.QPoint(recthorz+30,-1*rectvert))
        #outhdr = QtCore.QRectF(QtCore.QPoint(70,-90),QtCore.QPoint(90,-80))
        f.setUnderline(True)
        p.setFont(f)
        p.drawText(inphdr,QtCore.Qt.AlignCenter,'inputs')
        p.drawText(outhdr,QtCore.Qt.AlignCenter,'outputs')
        f.setUnderline(False)
        p.setFont(f)
        # Label the inputs
        n_inp = len(self.op.inputs)
        if n_inp > 0:
            ispc = rectvert/n_inp 
        else:
            ispc = rectvert
        vcrd = -1*rectvert+ispc
        for name in self.op.inputs.keys():
            il = self.op.input_locator[name]
            rec = QtCore.QRectF(QtCore.QPoint(-1*(recthorz-10),vcrd-5),QtCore.QPoint(0,vcrd+5))
            p.drawText(rec,QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter,name)
            p.drawLine(QtCore.QPoint(-1*(recthorz-5),vcrd),QtCore.QPoint(-1*(recthorz+10),vcrd))
            p.drawLine(QtCore.QPoint(-1*(recthorz+10),vcrd-10),QtCore.QPoint(-1*(recthorz+10),vcrd+10))
            ilrec = QtCore.QRectF(QtCore.QPoint(-100,vcrd-10),QtCore.QPoint(-1*(recthorz+12),vcrd+10))
            #p.drawText(ilrec,QtCore.Qt.AlignRight|QtCore.Qt.AlignVCenter,#|QtCore.Qt.TextWordWrap,
            vcrd += 2*ispc
        # Label the outputs
        n_out = len(self.op.outputs)
        ispc = 2*rectvert/(2*n_out)
        vcrd = -1*rectvert+ispc
        for name,val in self.op.outputs.items():
            rec = QtCore.QRectF(QtCore.QPoint(0,vcrd-5),QtCore.QPoint(recthorz-10,vcrd+5))
            p.drawText(rec,QtCore.Qt.AlignRight|QtCore.Qt.AlignVCenter,name)
            p.drawLine(QtCore.QPoint(recthorz-5,vcrd),QtCore.QPoint(recthorz+10,vcrd))
            p.drawLine(QtCore.QPoint(recthorz+10,vcrd-10),QtCore.QPoint(recthorz+10,vcrd+10))
            outrec = QtCore.QRectF(QtCore.QPoint(recthorz+12,vcrd-10),QtCore.QPoint(100,vcrd+10))
            #p.drawText(outrec,QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter,type(val).__name__)
            #p.drawText(outrec,QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter,str(val))
            #|QtCore.Qt.TextWordWrap,str(val))
            vcrd += 2*ispc




import unittest

class TestAPI(unittest.TestCase):

    def __init__(self,methodname,paw=None):
        super(TestAPI,self).__init__(methodname)
        self.paw = paw

    def test_start_stop(self):
        # smoke testing paws api
        import paws.api
        paw = paws.api.start()
        self.assertIsInstance(paw,paws.api.PawsAPI)

    def test_add_wf(self):
        from paws.core.workflow.Workflow import Workflow
        self.paw.add_wf('test')
        self.assertIsInstance(self.paw.get_wf('test'),Workflow)

    def test_add_op(self):
        from paws.core.operations.Operation import Operation
        self.paw.add_op('listprimes','TESTS.ListPrimes','test')
        self.assertIsInstance(self.paw.get_op('listprimes','test'),Operation)

    def test_execute(self):
        self.paw.execute('test')
        self.assertIsNotNone(self.paw.get_output('listprimes','primes_list','test'))
        print(self.paw.get_output('listprimes','primes_list'))

    def test_save(self):
        import os
        from paws.core import pawstools
        self.paw.save_to_wfl(os.path.join(pawstools.paws_scratch_dir,'test.wfl'))

    def test_load(self):
        import os
        from paws.core import pawstools
        self.paw.load_from_wfl(os.path.join(pawstools.paws_scratch_dir,'test.wfl'))

if __name__ == '__main__':
    unittest.main()




import pkgutil

def load_plugins(path_,pkg):
    p_names = []
    # pkgutil.iter_modules returns module_loader, module_name, ispkg forall modules in path
    mods = pkgutil.iter_modules(path_)
    mods = [mod for mod in mods if mod[1] not in ['__init__','PawsPlugin','PluginManager']]
    for modloader, modname, ispkg in mods:
        p_names.append(modname)
    return p_names 

plugin_name_list = load_plugins(__path__,__name__)
    
